"""
Commands related to torrent files
"""

from .... import jobs, trackers, utils
from . import argtypes
from .base import CommandBase


class torrent_add(CommandBase):
    """Add torrent file to BitTorrent client"""

    names = ('torrent-add', 'ta')

    argument_definitions = {
        'CLIENT': {
            'type': argtypes.client,
            'help': ('Case-insensitive BitTorrent client name\n'
                     'Supported clients: ' + ', '.join(utils.btclients.client_names())),
        },
        'TORRENT': {
            'nargs': '+',
            'help': 'Path to torrent file',
        },
        ('--download-path', '-p'): {
            'help': "Parent directory of the torrent's content",
            'metavar': 'DIRECTORY',
        },
    }

    @utils.cached_property
    def jobs(self):
        return (
            jobs.torrent.AddTorrentJob(
                home_directory=self.home_directory,
                cache_directory=self.cache_directory,
                ignore_cache=self.args.ignore_cache,
                client=utils.btclients.client(
                    name=self.args.CLIENT,
                    **self.config['clients'][self.args.CLIENT],
                ),
                download_path=self.args.download_path,
                enqueue=self.args.TORRENT,
            ),
        )


class torrent_create(CommandBase):
    """Create torrent file and optionally add or copy it"""

    names = ('torrent-create', 'tc')

    argument_definitions = {}

    subcommands = {
        tracker.name: {
            # Default arguments for all tackers
            **{
                'CONTENT': {
                    'type': argtypes.content,
                    'help': 'Path to release content',
                },
                ('--add-to', '-a'): {
                    'type': argtypes.client,
                    'metavar': 'CLIENT',
                    'help': ('Case-insensitive BitTorrent client name\n'
                             'Supported clients: ' + ', '.join(utils.btclients.client_names())),
                },
                ('--copy-to', '-c'): {
                    'metavar': 'PATH',
                    'help': 'Copy the created torrent to PATH (file or directory)',
                },
            },
            # Custom arguments defined by tracker
            **tracker.argument_definitions,
        }
        for tracker in trackers.trackers()
    }

    @utils.cached_property
    def tracker_name(self):
        return self.args.subcommand.lower()

    @utils.cached_property
    def create_torrent_job(self):
        return jobs.torrent.CreateTorrentJob(
            home_directory=self.home_directory,
            cache_directory=self.cache_directory,
            ignore_cache=self.args.ignore_cache,
            content_path=self.args.CONTENT,
            tracker=trackers.tracker(
                name=self.tracker_name,
                config=self.config['trackers'][self.tracker_name],
                cli_args=self.args,
            ),
        )

    @utils.cached_property
    def add_torrent_job(self):
        if self.args.add_to:
            add_torrent_job = jobs.torrent.AddTorrentJob(
                home_directory=self.home_directory,
                cache_directory=self.cache_directory,
                ignore_cache=self.args.ignore_cache,
                client=utils.btclients.client(
                    name=self.args.add_to,
                    **self.config['clients'][self.args.add_to],
                ),
                download_path=utils.fs.dirname(self.args.CONTENT),
            )
            # Pass CreateTorrentJob output to AddTorrentJob input.
            self.create_torrent_job.signal.register('output', add_torrent_job.enqueue)
            # Tell AddTorrentJob to finish the current upload and then finish.
            self.create_torrent_job.signal.register('finished', lambda _: add_torrent_job.finalize())
            return add_torrent_job

    @utils.cached_property
    def copy_torrent_job(self):
        if self.args.copy_to:
            copy_torrent_job = jobs.torrent.CopyTorrentJob(
                home_directory=self.home_directory,
                cache_directory=self.cache_directory,
                ignore_cache=self.args.ignore_cache,
                destination=self.args.copy_to,
            )
            # Pass CreateTorrentJob output to CopyTorrentJob input.
            self.create_torrent_job.signal.register('output', copy_torrent_job.enqueue)
            # Tell CopyTorrentJob to finish when CreateTorrentJob is done.
            self.create_torrent_job.signal.register('finished', lambda _: copy_torrent_job.finalize())
            return copy_torrent_job

    @utils.cached_property
    def jobs(self):
        return (
            self.create_torrent_job,
            self.add_torrent_job,
            self.copy_torrent_job,
        )



"""
Swiss Army knife
"""

import collections
import importlib
import inspect
import itertools
import os
import types as _types


def os_family():
    """
    Return "windows" or "unix"
    """
    return 'windows' if os.name == 'nt' else 'unix'


try:
    from functools import cached_property
except ImportError:
    import functools

    # https://github.com/jackmaney/lazy-property/
    class cached_property():
        """Property that replaces itself with its return value on first access"""

        def __init__(self, fget):
            self.fget = fget
            functools.update_wrapper(self, fget)
            self.fset = None

        def __get__(self, obj, cls):
            if obj is None:
                return self
            else:
                value = self.fget(obj)
                setattr(obj, self.fget.__name__, value)
                return value


# https://github.com/tensorflow/tensorflow/blob/master/tensorflow/python/util/lazy_loader.py
class LazyModule(_types.ModuleType):
    """
    Lazily import module to decrease execution time

    :param str module: Name of the module
    :param mapping namespace: Usually the return value of `globals()`
    :param str name: Name of the module in `namespace`; defaults to `module`
    """

    def __init__(self, module, namespace, name=None):
        self._module = module
        self._namespace = namespace
        self._name = name or module
        super().__init__(module)

    def _load(self):
        # Import the target module and insert it into the parent's namespace
        module = importlib.import_module(self.__name__)
        self._namespace[self._name] = module

        # Update this object's dict so that if someone keeps a reference to the
        # LazyLoader, lookups are efficient (__getattr__ is only called on
        # lookups that fail).
        self.__dict__.update(module.__dict__)

        return module

    def __getattr__(self, item):
        module = self._load()
        return getattr(module, item)

    def __dir__(self):
        module = self._load()
        return dir(module)


def submodules(package):
    """
    Return list of submodules and subpackages in `package`

    :param str package: Fully qualified name of parent package,
        e.g. "upsies.utils.btclients"
    """
    # Get absolute path to parent directory of top-level package
    own_path = os.path.dirname(__file__)
    rel_path = __package__.replace('.', '/')
    assert own_path.endswith(rel_path)
    project_path = own_path[:-len(rel_path)]

    # Add path to given package
    package_path = os.path.join(project_path, package.replace('.', '/'))

    # Find and import public submodules
    submods = []
    for name in os.listdir(package_path):
        if not name.startswith('_'):
            if name.endswith('.py'):
                name = name[:-3]
            if '.' not in name:
                submods.append(
                    importlib.import_module(name=f'.{name}', package=package)
                )
    return submods


def subclasses(basecls, modules):
    """
    Find subclasses in modules

    :param type basecls: Class that all returned classes are a subclass of
    :param modules: Modules to search
    :type modules: list of module objects
    """
    subclses = set()
    for mod in modules:
        for name, member in inspect.getmembers(mod):
            if (member is not basecls and
                isinstance(member, type) and
                issubclass(member, basecls)):
                subclses.add(member)
    return subclses


def closest_number(n, ns, max=None, default=0):
    """
    Return the number from `ns` that is closest to `n`

    :param n: Given number
    :param ns: Sequence of allowed numbers
    :param max: Remove any item from `ns` that is larger than `max`
    :param default: Return value in case `ns` is empty
    """
    if max is not None:
        ns_ = tuple(n_ for n_ in ns if n_ <= max)
        if not ns_:
            raise ValueError(f'No number equal to or below {max}: {ns}')
    else:
        ns_ = ns
    return min(ns_, key=lambda x: abs(x - n), default=default)


class CaseInsensitiveString(str):
    """String that ignores case when compared or sorted"""

    def __hash__(self):
        return hash(self.casefold())

    def __eq__(self, other):
        if not isinstance(other, str):
            return NotImplemented
        else:
            return self.casefold() == other.casefold()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        return self.casefold() < other.casefold()

    def __le__(self, other):
        return self.casefold() <= other.casefold()

    def __gt__(self, other):
        return self.casefold() > other.casefold()

    def __ge__(self, other):
        return self.casefold() >= other.casefold()


def is_sequence(obj):
    """Return whether `obj` is a sequence and not a string"""
    return (isinstance(obj, collections.abc.Sequence)
            and not isinstance(obj, str))


def merge_dicts(a, b, path=()):
    """
    Merge nested dictionaries `a` and `b` into new dictionary with same
    structure
    """
    keys = itertools.chain(a, b)
    merged = {}
    for key in keys:
        if isinstance(a.get(key), dict) and isinstance(b.get(key), dict):
            merged[key] = merge_dicts(a[key], b[key], path + (key,))
        elif key in b:
            # Value from b takes precedence
            merged[key] = b[key]
        elif key in a:
            # Value from a is default
            merged[key] = a[key]
    return merged


def deduplicate(l, key=None):
    """
    Return sequence `l` with all duplicate items removed while maintaining the
    original order

    :param key: Callable that gets each item and returns a hashable identifier
        for that item
    """
    if key is None:
        key = lambda k: k
    seen_keys = set()
    deduped = []
    for item in l:
        k = key(item)
        if k not in seen_keys:
            seen_keys.add(k)
            deduped.append(item)
    return deduped


from . import (browser, btclients, configfiles, daemon, fs, html, http, image,
               imghosts, iso, release, scene, signal, string, subproc,
               timestamp, torrent, types, video, webdbs)



from ....utils import cached_property, release
from .. import widgets
from . import JobWidgetBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class ReleaseNameJobWidget(JobWidgetBase):
    def setup(self):
        self._input = widgets.InputField(
            text='Loading...',
            style='class:dialog.text',
            on_accepted=self._handle_approved_release_name,
            read_only=True,
        )
        self.job.signal.register('release_name_updated', self._set_input)

    def _set_input(self, release_name):
        self._input.read_only = False
        self._input.text = self._get_release_name_string(release_name)
        self.invalidate()

    def _get_release_name_string(self, release_name):
        # TODO: Use custom format specified by some attribute on self.job.
        return release_name.format()

    def _handle_approved_release_name(self, _):
        _log.debug('Approved release name: %r', self._input.text)
        release_name = release.ReleaseName(self._input.text)
        if release_name.is_complete:
            self.job.release_name_selected(self._input.text)
        else:
            self._set_input(release_name)

    @cached_property
    def runtime_widget(self):
        return self._input



"""
Managing callbacks
"""

import collections

import logging  # isort:skip
_log = logging.getLogger(__name__)


class Signal:
    """Simple callback registry"""

    def __init__(self, *signals):
        self._signals = {}
        self._emissions = []
        self._record_signals = []
        for signal in signals:
            self.add(signal)

    def add(self, signal, record=False):
        """
        Create new `signal`

        :param hashable signal: Any hashable object
        :param bool record: Whether emissions of this signal are recorded in
            :attr:`emissions`

        :raise TypeError: if `signal` is not hashable
        :raise RuntimeError: if `signal` has been added previously
        """
        if not isinstance(signal, collections.abc.Hashable):
            raise TypeError(f'Unhashable signal: {signal!r}')
        elif signal in self._signals:
            raise RuntimeError(f'Signal already added: {signal!r}')
        else:
            self._signals[signal] = []
            if record:
                self.record(signal)

    @property
    def signals(self):
        """Mutable dictionary of signals mapped to a lists of callbacks"""
        return self._signals

    def record(self, signal):
        """Record emissions of `signal` in :attr:`emissions`"""
        if signal not in self._signals:
            raise ValueError(f'Unknown signal: {signal!r}')
        else:
            self._record_signals.append(signal)

    @property
    def recording(self):
        """class:`list` of signals that are recorded"""
        return self._record_signals

    def register(self, signal, callback):
        """
        Call `callback` when `signal` is emited

        :param hashable signal: Previously added signal
        :param callable callback: Any callback. The signature depends on the caller of
            :meth:`emit`.

        :raise ValueError: if `signal` was not added first
        :raise TypeError: if `callback` is not callable
        """
        if signal not in self._signals:
            raise ValueError(f'Unknown signal: {signal!r}')
        elif not callable(callback):
            raise TypeError(f'Not a callable: {callback!r}')
        else:
            self._signals[signal].append(callback)

    def emit(self, signal, *args, **kwargs):
        """
        Call callbacks that are registered to `signal`

        :param hashable signal: Previously added signal

        Any other arguments are passed on to the callbacks.
        """
        for callback in self._signals[signal]:
            callback(*args, **kwargs)
        if signal in self._record_signals:
            self._emissions.append((signal, {'args': args, 'kwargs': kwargs}))

    @property
    def emissions(self):
        """
        Sequence of recorded :meth:`emit` calls

        Each call is stored as a tuple like this::

            (<signal>, {"args": <positional arguments>,
                        "kwargs": <keyword arguments>})
        """
        return tuple(self._emissions)

    def replay(self, emissions):
        """:meth:`emit` previously recorded :attr:`emissions`"""
        for signal, payload in emissions:
            self.emit(signal, *payload['args'], **payload['kwargs'])



"""
Dummy client API for testing and debugging
"""

import asyncio

from .base import ClientApiBase


class DummyClientApi(ClientApiBase):
    """Dummy client API for testing and debugging"""

    name = 'dummy'

    async def add_torrent(self, torrent_path, download_path=None):
        """Pretend to add `torrent_path`"""
        # Raise exception if file doesn't exist
        self.read_torrent_file(torrent_path)
        await asyncio.sleep(2)
        return 'DE4DB33F'



"""
CLI argument types

All types return normalized values and raise ValueError for invalid values.

A custom error message can be provided by raising
:class:`argparse.ArgumentTypeError`.
"""

import argparse
import os

from .... import defaults, errors, trackers, utils


def client(value):
    """Name of a BitTorrent client from :mod:`~.utils.btclients`"""
    if value in utils.btclients.client_names():
        return value.lower()
    else:
        raise ValueError(f'Unsupported client: {value}')


def content(value):
    """Existing path to release file(s)"""
    path = release(value)
    if not os.path.exists(path):
        raise argparse.ArgumentTypeError(f'No such file or directory: {value}')
    else:
        return path


def imghost(value):
    """Name of a image hosting service from :mod:`~.utils.imghosts`"""
    if value in utils.imghosts.imghost_names():
        return value.lower()
    else:
        raise ValueError(f'Unsupported image hosting service: {value}')


def integer(value):
    """Natural number (:class:`float` is rounded)"""
    try:
        return int(float(value))
    except (ValueError, TypeError):
        raise ValueError(f'Not an integer: {value!r}')


def option(value):
    """Name of a configuration option"""
    if value in defaults.option_paths():
        return value.lower()
    else:
        raise ValueError(f'Unknown option: {value}')


def release(value):
    """Same as :func:`content`, but doesn't have to exist"""
    path = str(value)
    try:
        utils.scene.assert_not_abbreviated_filename(path)
    except errors.SceneError as e:
        raise argparse.ArgumentTypeError(e)
    else:
        return path


def scenedb(value):
    """Name of a scene release database from :mod:`~.utils.scene`"""
    if value in utils.scene.scenedb_names():
        return value.lower()
    else:
        raise ValueError(f'Unsupported scene release database: {value}')


def timestamp(value):
    """See :func:`.timestamp.parse`"""
    try:
        return utils.timestamp.parse(value)
    except TypeError as e:
        raise ValueError(e)


def tracker(value):
    """Name of a tracker from :mod:`~.trackers`"""
    if value in trackers.tracker_names():
        return value.lower()
    else:
        raise ValueError(f'Unsupported tracker: {value}')


def webdb(value):
    """Name of a movie/series database from :mod:`~.webdbs`"""
    if value in utils.webdbs.webdb_names():
        return value.lower()
    else:
        raise ValueError(f'Unsupported database: {value}')



"""
Create torrent file
"""

import time
from os.path import exists as _path_exists

from .. import __project_name__, __version__, errors
from . import LazyModule

import logging  # isort:skip
_log = logging.getLogger(__name__)

torf = LazyModule(module='torf', namespace=globals())


def create(*, content_path, announce, torrent_path,
           init_callback, progress_callback,
           overwrite=False, source=None, exclude=()):
    """
    Generate and write torrent file

    :param str content_path: Path to the torrent's payload
    :param str announce: Announce URL
    :param str torrent_path: Path of the generated torrent file
    :param str init_callback: Callable that is called once before torrent
        generation commences. It gets `content_path` as a tree where a node is a
        tuple in which the first item is the directory name and the second item
        is a list of `(file_name, file_size)` tuples.

        Example:

        .. code::

           ('Parent',
               ('Foo', (
                   ('Picture.jpg', 82489),
                   ('Music.mp3', 5315672),
                   ('More files', (
                       ('This.txt', 57734),
                       ('And that.txt', 184),
                       ('Also some of this.txt', 88433),
                   )),
               )),
               ('Bar', (
                   ('Yee.mp4', 288489392),
                   ('Yah.mkv', 3883247384),
               )),
           )
    :param str progress_callback: Callable that gets the progress as a number
        between 0 and 100. Torrent creation is cancelled if `progress_callback`
        returns `True` or any other truthy value.
    :param bool overwrite: Whether to overwrite `torrent_path` if it exists
    :param str source: Value of the "source" field in the torrent or `None` to
        leave it out
    :param exclude: Sequence of regular expressions; matching files are not
        included in the torrent

    :raise TorrentError: if anything goes wrong

    :return: `torrent_path`
    """
    if not announce:
        raise errors.TorrentError('Announce URL is empty')

    def cb(torrent, filepath, pieces_done, pieces_total):
        return progress_callback(pieces_done / pieces_total * 100)

    if not overwrite and _path_exists(torrent_path):
        _log.debug('Torrent file already exists: %r', torrent_path)
        return torrent_path
    else:
        try:
            torrent = torf.Torrent(
                path=content_path,
                exclude_regexs=exclude,
                trackers=((announce,),),
                private=True,
                source=source,
                created_by=f'{__project_name__} {__version__}',
                creation_date=time.time(),
            )
            init_callback(_make_file_tree(torrent.filetree))
            success = torrent.generate(callback=cb, interval=0.5)
        except torf.TorfError as e:
            raise errors.TorrentError(e)

        if success:
            try:
                torrent.write(torrent_path, overwrite=overwrite)
            except torf.TorfError as e:
                raise errors.TorrentError(e)
            else:
                return torrent_path


def _make_file_tree(tree):
    files = []
    for name,file in tree.items():
        if isinstance(file, dict):
            subtree = _make_file_tree(file)
            files.append((name, subtree))
        else:
            files.append((name, file.size))
    return tuple(files)



"""
String formatting and parsing
"""

import math

_byte_units = {
    'PiB': 1024 ** 5,
    'TiB': 1024 ** 4,
    'GiB': 1024 ** 3,
    'MiB': 1024 ** 2,
    'KiB': 1024,
}

def pretty_bytes(b):
    '''Return `b` as human-readable bytes, e.g. "24.3 MiB"'''
    for unit, bytes in _byte_units.items():
        if b >= bytes:
            return f'{b / bytes:.2f} {unit}'
    return f'{int(b)} B'


def star_rating(rating, max_rating=10):
    '''Return star rating string with the characters "★", "⯪" and "☆"'''
    rating = min(max_rating, max(0, rating))
    left = '\u2605' * math.floor(rating)
    if rating >= max_rating:
        middle = ''
    # Avoid floating point precision issues by rounding to 1 digit after comma
    elif round(rating % 1, 1) <= 0.3:
        middle = '\u2606'  # Empty star
    elif round(rating % 1, 1) < 0.7:
        middle = '\u2bea'  # Half star
    else:
        middle = '\u2605'  # Full star
    right = '\u2606' * (math.ceil(max_rating - rating) - 1)
    return f'{left}{middle}{right}'



"""
HTTP methods with caching
"""

import asyncio
import collections
import hashlib
import json
import os

import httpx

from .. import __project_name__, __version__, constants, errors
from . import fs

import logging  # isort:skip
_log = logging.getLogger(__name__)

_default_headers = {
    'Accept-Language': 'en-US,en;q=0.5',
    'User-Agent': f'{__project_name__}/{__version__}',
}

# Use one Lock per requested url + parameters so we can make multiple identical
# requests concurrently without bugging the server.
_request_locks = collections.defaultdict(lambda: asyncio.Lock())

_client = httpx.AsyncClient(
    timeout=60,
)


cache_directory = None
"""
Where to store cached requests

If this is set to a falsy value, default to :attr:`~.constants.CACHE_DIRPATH`.
"""


def close():
    """Close the client session"""
    _log.debug('Closing client: %r', _client)
    asyncio.get_event_loop().run_until_complete(_client.aclose())
    _log.debug('Closed client: %r', _client)


async def get(url, headers={}, params={}, auth=None,
              cache=False, user_agent=False, allow_redirects=True):
    """
    Perform HTTP GET request

    :param str url: URL to request
    :param dict headers: Custom headers (added to default headers)
    :param dict params: Query arguments, e.g. `{"k": "v"} → http://foo/bar?k=v`
    :param auth: Basic access authentication; sequence of <username> and
        <password> or `None`
    :param bool cache: Whether to use cached response if available
    :param bool user_agent: Whether to send the User-Agent header
    :param bool allow_redirects: Whether to follow redirects

    :return: Response text
    :rtype: Response
    :raise RequestError: if the request fails for any expected reason
    """
    return await _request(
        method='GET',
        url=url,
        headers=headers,
        params=params,
        auth=auth,
        cache=cache,
        user_agent=user_agent,
        allow_redirects=allow_redirects,
    )

async def post(url, headers={}, data={}, files={}, auth=None,
               cache=False, user_agent=False, allow_redirects=True):
    """
    Perform HTTP POST request

    :param str url: URL to request
    :param dict headers: Custom headers (added to default headers)
    :param dict data: Data to send as application/x-www-form-urlencoded
    :param dict files: Files to send as multipart/form-data
    :param auth: Basic access authentication; sequence of <username> and
        <password> or `None`
    :param bool cache: Whether to use cached response if available
    :param bool user_agent: Whether to send the User-Agent header
    :param bool allow_redirects: Whether to follow redirects

    :return: Response text
    :rtype: Response
    :raise RequestError: if the request fails for any expected reason
    """
    return await _request(
        method='POST',
        url=url,
        headers=headers,
        data=data,
        files=files,
        auth=auth,
        cache=cache,
        user_agent=user_agent,
        allow_redirects=allow_redirects,
    )

async def download(url, filepath, *args, **kwargs):
    """
    Write downloaded data to file

    :param url: Where to download the data from
    :param filepath: Where to save the downloaded data

    Any other arguments are passed to :func:`get`, except for `cache`, which is
    always `False`.

    If `filepath` exists, no request is made.

    :raise RequestError: if anything goes wrong
    :return: `filepath`
    """
    kwargs['cache'] = False
    if not os.path.exists(filepath):
        _log.debug('Downloading %r to %r', url, filepath)
        response = await get(url, *args, **kwargs)
        try:
            with open(filepath, 'wb') as f:
                f.write(response.bytes)
        except OSError as e:
            if e.strerror:
                raise errors.RequestError(f'Unable to write {filepath}: {e.strerror}')
            else:
                raise errors.RequestError(f'Unable to write {filepath}: {e}')
    else:
        _log.debug('Already downloaded %r to %r', url, filepath)
    return filepath


class Result(str):
    """
    Response to an HTTP request

    This is a subclass of :class:`str` with additional attributes and methods.
    """

    def __new__(cls, text, bytes, headers={}, status_code=None):
        obj = super().__new__(cls, text)
        obj._bytes = bytes
        obj._headers = headers
        obj._status_code = status_code
        return obj

    @property
    def headers(self):
        """HTTP headers"""
        return self._headers

    @property
    def status_code(self):
        """HTTP status code"""
        return self._status_code

    @property
    def bytes(self):
        """Response data as :class:`bytes`"""
        return self._bytes

    def json(self):
        """
        Parse the response text as JSON

        :return: JSON as a `dict`
        :raise Requesterror: if parsing fails
        """
        try:
            return json.loads(self)
        except ValueError as e:
            raise errors.RequestError(f'Malformed JSON: {str(self)}: {e}')

    def __repr__(self):
        return (f'{type(self).__name__}('
                f'text={str(self)!r}, '
                f'bytes={self.bytes!r}, '
                f'headers={self.headers!r}, '
                f'status_code={self.status_code!r})')


async def _request(method, url, headers={}, params={}, data={}, files={},
                   allow_redirects=True, cache=False, auth=None, user_agent=False):
    assert isinstance(user_agent, bool)
    if method.upper() not in ('GET', 'POST'):
        raise ValueError(f'Invalid method: {method}')

    if isinstance(data, (bytes, str)):
        payload = {'content': data}
    else:
        payload = {'data': data}

    headers = {**_default_headers, **headers}
    request = _client.build_request(
        method=str(method),
        headers=headers,
        url=str(url),
        params=params,
        files=_open_files(files),
        **payload,
    )

    if not user_agent:
        del request.headers['User-Agent']

    # Block when requesting the same URL simultaneously
    request_lock_key = (request.url, await request.aread())
    # _log.debug('Request lock key: %r', request_lock_key)
    request_lock = _request_locks[request_lock_key]
    async with request_lock:
        if cache:
            cache_file = _cache_file(method, url, params)
            result = _from_cache(cache_file)
            if result is not None:
                return result

        _log.debug('%s: %r: %r', method, url, params)
        try:
            response = await _client.send(
                request=request,
                auth=auth,
                allow_redirects=allow_redirects,
            )
            try:
                response.raise_for_status()
            except httpx.HTTPStatusError:
                raise errors.RequestError(
                    f'{url}: {response.text}',
                    headers=response.headers,
                    status_code=response.status_code,
                )
        except httpx.TimeoutException:
            raise errors.RequestError(f'{url}: Timeout')
        except httpx.HTTPError as e:
            _log.debug(f'Unexpected HTTP error: {e!r}')
            raise errors.RequestError(f'{url}: {e}')
        else:
            if cache:
                cache_file = _cache_file(method, url, params)
                _to_cache(cache_file, response.content)
            return Result(
                text=response.text,
                bytes=response.content,
                headers=response.headers,
                status_code=response.status_code,
            )


def _open_files(files):
    """
    Open files for upload

    :param files: Dictionary that maps field names to file paths or to 2-tuples
        of file paths and MIME types

        Example:

        >>> _open_files({
        >>>     "image": "path/to/foo.jpg",
        >>>     "document": ("path/to/bar", "text/plain"),
        >>> })

    :return: See https://www.python-httpx.org/advanced/#multipart-file-encoding

    :raise RequestError: if a file cannot be opened
    """
    opened = {}
    for fieldname, fileinfo in files.items():
        if isinstance(fileinfo, str):
            opened[fieldname] = (
                os.path.basename(fileinfo),     # File name
                _get_fileobj(fileinfo),         # File handle
            )
        elif isinstance(fileinfo, collections.abc.Sequence) and len(fileinfo) == 2:
            opened[fieldname] = (
                os.path.basename(fileinfo[0]),  # File name
                _get_fileobj(fileinfo[0]),      # File handle
                fileinfo[1],                    # MIME type
            )
        else:
            raise RuntimeError(f'Invalid file: {fileinfo}')
    return opened


def _get_fileobj(filepath):
    """
    Open `filepath` and return the file object

    :raise RequestError: if OSError is raised by `open`
    """
    try:
        return open(filepath, 'rb')
    except OSError as e:
        if e.strerror:
            msg = e.strerror
        else:
            msg = 'Failed to open'
        raise errors.RequestError(f'{filepath}: {msg}')


def _to_cache(cache_file, bytes):
    try:
        fs.mkdir(fs.dirname(cache_file))
        with open(cache_file, 'wb') as f:
            f.write(bytes)
    except OSError as e:
        raise RuntimeError(f'Unable to write cache file {cache_file}: {e}')


def _from_cache(cache_file):
    bytes = _read_bytes_from_file(cache_file)
    text = _read_string_from_file(cache_file)
    if bytes and text:
        return Result(text=text, bytes=bytes)


def _read_string_from_file(filepath):
    bytes = _read_bytes_from_file(filepath)
    if bytes:
        return str(bytes, encoding='utf-8', errors='replace')


def _read_bytes_from_file(filepath):
    try:
        with open(filepath, 'rb') as f:
            return f.read()
    except OSError:
        pass


def _cache_file(method, url, params={}):
    def make_filename(method, url, params_str):
        if params_str:
            filename = f'{method.upper()}.{url}?{params_str}'
        else:
            filename = f'{method.upper()}.{url}'
        return filename.replace('/', '_').replace(' ', '+')

    if params:
        params_str = '&'.join((f'{k}={v}' for k,v in params.items()))
        if len(make_filename(method, url, params_str)) > 250:
            params_str = f'[HASH:{_semantic_hash(params)}]'
    else:
        params_str = ''

    return os.path.join(
        cache_directory or constants.CACHE_DIRPATH,
        make_filename(method, url, params_str),
    )


def _semantic_hash(obj):
    """
    Return hash for `obj` that stays the same between sessions

    https://github.com/schollii/sandals/blob/master/json_sem_hash.py
    """
    def sorted_dict_str(obj):
        if isinstance(obj, collections.abc.Mapping):
            return {k: sorted_dict_str(obj[k]) for k in sorted(obj.keys())}
        elif isinstance(obj, str):
            return str(obj)
        elif isinstance(obj, collections.abc.Sequence):
            return [sorted_dict_str(val) for val in obj]
        else:
            return str(obj)

    return hashlib.sha256(bytes(repr(sorted_dict_str(obj)), 'UTF-8')).hexdigest()



"""
Fixed values that do not change during runtime
"""

import os

from xdg.BaseDirectory import xdg_cache_home as XDG_CACHE_HOME
from xdg.BaseDirectory import xdg_config_home as XDG_CONFIG_HOME

CACHE_DIRPATH = os.path.join(XDG_CACHE_HOME, 'upsies')
"""Path cache directory"""

CONFIG_FILEPATH = os.path.join(XDG_CONFIG_HOME, 'upsies', 'config.ini')
"""Path to general configuration file"""

TRACKERS_FILEPATH = os.path.join(XDG_CONFIG_HOME, 'upsies', 'trackers.ini')
"""Path to trackers configuration file"""

CLIENTS_FILEPATH = os.path.join(XDG_CONFIG_HOME, 'upsies', 'clients.ini')
"""Path to BitTorrent clients configuration file"""

def _get_option_paths(cfg, parents=()):
    paths = []
    for k, v in cfg.items():
        k_parents = parents + (k,)
        if isinstance(v, dict):
            paths.extend(_get_option_paths(v, parents=k_parents))
        else:
            paths.append('.'.join(k_parents))
    return tuple(sorted(paths))

VIDEO_FILE_EXTENSIONS = ('mkv', 'mp4', 'ts', 'avi')
"""Sequence of file extensions to recognize video files"""

GUESSIT_OPTIONS = {
    'expected_title': [
        'The Collector',
        'xXx',
    ],
    'advanced_config': {
        # https://github.com/guessit-io/guessit/pull/678
        'audio_codec': {
            'audio_channels': {
                '1.0': [
                    '1ch',
                    'mono',
                    're:(1[\\W_]0(?:ch)?)(?=[^\\d]|$)',
                ],
            },
        },
    },
}
"""`options` argument for :meth:`.GuessItApi.guessit`"""



"""
Create screenshots from video file and optionally upload them
"""

from .... import jobs, utils
from . import argtypes
from .base import CommandBase


class screenshots(CommandBase):
    """Create screenshots from video file and optionally upload them"""

    names = ('screenshots', 'ss')

    argument_definitions = {
        'CONTENT': {
            'type': argtypes.content,
            'help': 'Path to release content',
        },
        ('--timestamps', '-t'): {
            'nargs': '+',
            'default': (),
            'type': argtypes.timestamp,
            'metavar': 'TIMESTAMP',
            'help': 'Space-separated list of [[HH:]MM:]SS strings',
        },
        ('--number', '-n'): {
            'type': argtypes.integer,
            'help': 'How many screenshots to make in total',
            'default': 0,
        },
        ('--upload-to', '-u'): {
            'type': argtypes.imghost,
            'metavar': 'IMAGEHOST',
            'help': ('Case-insensitive name of image hosting service\n'
                     'Supported services: ' + ', '.join(utils.imghosts.imghost_names())),
        },
    }

    @utils.cached_property
    def screenshots_job(self):
        return jobs.screenshots.ScreenshotsJob(
            home_directory=self.home_directory,
            cache_directory=self.cache_directory,
            ignore_cache=self.args.ignore_cache,
            content_path=self.args.CONTENT,
            timestamps=self.args.timestamps,
            count=self.args.number,
        )

    @utils.cached_property
    def upload_screenshots_job(self):
        if self.args.upload_to:
            imghost_job = jobs.imghost.ImageHostJob(
                home_directory=self.home_directory,
                cache_directory=self.cache_directory,
                ignore_cache=self.args.ignore_cache,
                imghost=utils.imghosts.imghost(
                    name=self.args.upload_to,
                    **self.config['imghosts'][self.args.upload_to],
                ),
            )
            # Timestamps are calculated in a subprocess, we have to wait for
            # that until we can set the number of expected screenhots.
            self.screenshots_job.signal.register(
                'timestamps',
                lambda timestamps: imghost_job.set_images_total(len(timestamps)),
            )
            # Pass ScreenshotsJob's output to ImageHostJob input.
            self.screenshots_job.signal.register('output', imghost_job.enqueue)
            # Tell imghost_job to finish the current upload and then finish.
            self.screenshots_job.signal.register('finished', lambda _: imghost_job.finalize())
            return imghost_job

    @utils.cached_property
    def jobs(self):
        return (
            self.screenshots_job,
            self.upload_screenshots_job,
        )



"""
HTML parsing
"""

from .. import errors
from . import LazyModule

bs4 = LazyModule(module='bs4', namespace=globals())


def parse(string):
    """
    Return :class:`~.bs4.BeautifulSoup` instance

    :param string: HTML document

    :raise ContentError: if `string` is invalid HTML
    """
    try:
        return bs4.BeautifulSoup(str(string), features='html.parser')
    except Exception as e:
        raise errors.ContentError(f'Invalid HTML: {e}')


def dump(html, filepath):
    """
    Write `html` to `filepath` for debugging

    :param html: String or :class:`~.bs4.BeautifulSoup` instance
    """
    with open(filepath, 'w') as f:
        if isinstance(html, bs4.BeautifulSoup):
            f.write(html.prettify())
        else:
            f.write(parse(str(html)).prettify())



"""
API for themoviedb.org
"""

import functools
import re

from .. import html, http
from ..types import ReleaseType
from . import common
from .base import WebDbApiBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class TmdbApi(WebDbApiBase):
    """API for themoviedb.org"""

    name = 'tmdb'
    label = 'TMDb'
    _url_base = 'http://themoviedb.org'

    _soup_cache = {}

    async def _get_soup(self, path, params={}):
        cache_id = (path, tuple(sorted(params.items())))
        if cache_id in self._soup_cache:
            return self._soup_cache[cache_id]
        text = await http.get(
            url=f'{self._url_base}/{path.lstrip("/")}',
            params=params,
            cache=True,
        )
        self._soup_cache[cache_id] = html.parse(text)
        return self._soup_cache[cache_id]

    async def search(self, query):
        _log.debug('Searching TMDb for %s', query)
        if not query.title:
            return []

        params = {'query': query.title_normalized}
        if query.year is not None:
            params['query'] += f' y:{query.year}'

        if query.type is ReleaseType.movie:
            soup = await self._get_soup('/search/movie', params=params)
        elif query.type in (ReleaseType.season, ReleaseType.episode):
            soup = await self._get_soup('/search/tv', params=params)
        else:
            soup = await self._get_soup('/search', params=params)

        # When search for year, unmatching results are included but hidden.
        for tag in soup.find_all('div', class_='hide'):
            tag.clear()

        items = soup.find_all('div', class_='card')
        return [_TmdbSearchResult(soup=item, tmdb_api=self)
                for item in items]

    _person_url_path_regex = re.compile(r'(/person/\d+[-a-z]+)')

    def _get_persons(self, tag):
        a_tags = tag.find_all('a', href=self._person_url_path_regex)
        persons = []
        for a_tag in a_tags:
            if a_tag.string:
                name = a_tag.string.strip()
                url_path = self._person_url_path_regex.match(a_tag["href"]).group(1)
                url = f'{self._url_base.rstrip("/")}/{url_path.lstrip("/")}'
                persons.append(common.Person(name, url))
        return tuple(persons)

    async def cast(self, id):
        soup = await self._get_soup(id)
        cards = soup.select('.people > .card')
        cast = []
        for card in cards:
            cast.extend(self._get_persons(card))
        return tuple(cast)

    async def countries(self, id):
        raise NotImplementedError('Country lookup is not implemented for TMDb')

    async def creators(self, id):
        soup = await self._get_soup(id)
        profiles = soup.select('.people > .profile')
        creators = []
        for profile in profiles:
            if profile.find('p', text=re.compile(r'(?i:Creator)')):
                creators.extend(self._get_persons(profile))
        return tuple(creators)

    async def directors(self, id):
        soup = await self._get_soup(id)
        profiles = soup.select('.people > .profile')
        directors = []
        for profile in profiles:
            if profile.find('p', text=re.compile(r'(?i:Director)')):
                directors.extend(self._get_persons(profile))
        return tuple(directors)

    async def keywords(self, id):
        soup = await self._get_soup(id)
        keywords = soup.find(class_='keywords')
        if not keywords:
            return ()
        else:
            keywords = list(keywords.stripped_strings)
            if keywords[0] == 'Keywords':
                keywords.pop(0)
            if 'No keywords have been added.' in keywords:
                keywords.clear()
            return tuple(keywords)

    async def poster_url(self, id):
        soup = await self._get_soup(id)
        img_tag = soup.find('img', class_='poster')
        if img_tag:
            srcs = img_tag.get('data-srcset')
            if srcs:
                path = srcs.split()[0]
                return f'{self._url_base.rstrip("/")}/{path.lstrip("/")}'

    rating_min = 0.0
    rating_max = 100.0

    async def rating(self, id):
        soup = await self._get_soup(id)
        rating_tag = soup.find(class_='user_score_chart')
        if rating_tag:
            try:
                return float(rating_tag['data-percent'])
            except (ValueError, TypeError):
                pass

    _no_overview_texts = (
        "We don't have an overview",
        'No overview found.',
    )

    async def summary(self, id):
        soup = await self._get_soup(id)
        overview = ''.join(soup.find('div', class_='overview').stripped_strings)
        if any(text in overview for text in self._no_overview_texts):
            overview = ''
        return overview

    async def title_english(self, id):
        raise NotImplementedError('English title lookup is not implemented for TMDb')

    async def title_original(self, id):
        # TODO: Original title is supported
        # https://www.themoviedb.org/movie/3405-le-dolce-signore
        raise NotImplementedError('Original title lookup is not implemented for TMDb')

    async def type(self, id):
        raise NotImplementedError('Type lookup is not implemented for TMDb')

    async def url(self, id):
        return f'{self._url_base.rstrip("/")}/{id}'

    async def year(self, id):
        soup = await self._get_soup(id)
        year = ''.join(soup.find(class_='release_date').stripped_strings).strip('()')
        _log.debug(year)
        if len(year) == 4 and year.isdigit():
            return year
        else:
            return ''


class _TmdbSearchResult(common.SearchResult):
    def __init__(self, *, soup, tmdb_api):
        id = self._get_id(soup)
        return super().__init__(
            id=id,
            title=self._get_title(soup),
            type=self._get_type(soup),
            url=self._get_url(soup),
            year=self._get_year(soup),
            cast=functools.partial(tmdb_api.cast, id),
            countries=[],
            director='',
            keywords=functools.partial(tmdb_api.keywords, id),
            summary=functools.partial(tmdb_api.summary, id),
            title_english='',
            title_original='',
        )

    def _get_id(self, soup):
        href = soup.find('a', class_='result').get('href')
        return re.sub(r'^.*/((?:movie|tv)/[0-9]+).*?$', r'\1', href)

    def _get_url(self, soup):
        href = soup.find('a', class_='result').get('href')
        if href.startswith('http'):
            return href
        elif href.startswith('/'):
            return f'{TmdbApi._url_base}{href}'
        else:
            return f'{TmdbApi._url_base}/{href}'

    def _get_type(self, soup):
        href = soup.find('a', class_='result').get('href')
        tmdb_type = re.sub(r'(?:^|/)(\w+)/.*$', r'\1', href)
        if tmdb_type == 'movie':
            return ReleaseType.movie
        elif tmdb_type == 'tv':
            return ReleaseType.series
        else:
            return ReleaseType.unknown

    def _get_title(self, soup):
        header = soup.select('.result > h2')
        if header:
            return header[0].string
        else:
            return ''

    def _get_year(self, soup):
        release_date = soup.find(class_='release_date')
        if release_date:
            match = re.search(r'(\d{4})$', release_date.string)
            if match:
                return match.group(1)
        return ''



"""
Base class for image uploaders
"""

import abc
import json
import os

from ... import constants
from .. import fs
from . import common

import logging  # isort:skip
_log = logging.getLogger(__name__)


class ImageHostBase(abc.ABC):
    """
    Base class for image uploaders

    :param str cache_dir: Where to store URLs in JSON files; defaults to
        :attr:`.constants.CACHE_DIRPATH`
    """

    def __init__(self, cache_directory=None):
        self.cache_directory = cache_directory if cache_directory else constants.CACHE_DIRPATH

    @property
    @abc.abstractmethod
    def name(self):
        """Name of the image hosting service"""

    @property
    def cache_directory(self):
        """Path to directory where upload info is cached"""
        return self._cache_dir

    @cache_directory.setter
    def cache_directory(self, directory):
        self._cache_dir = directory

    async def upload(self, image_path, cache=True):
        """
        Upload image to gallery

        :param str image_path: Path to image file
        :param bool cache: Whether to attempt to get the image URL from cache or
            cache it

        :raise RequestError: if the upload fails

        :return: :class:`~.imghost.common.UploadedImage`
        """
        info = self._get_info_from_cache(image_path) if cache else {}
        if not info:
            info = await self._upload(image_path)
            _log.debug('Uploaded %r: %r', image_path, info)
            self._store_info_to_cache(image_path, info)
        if 'url' not in info:
            raise RuntimeError(f'Missing "url" key in {info}')
        return common.UploadedImage(**info)

    @abc.abstractmethod
    async def _upload(self, image_path):
        """
        Upload a single image

        :param str image_path: Path to an image file

        :return: Dictionary that must contain an "url" key
        """

    def _get_info_from_cache(self, image_path):
        cache_file = self._cache_file(image_path)
        if os.path.exists(cache_file):
            _log.debug('Already uploaded: %s', cache_file)
            try:
                with open(cache_file, 'r') as f:
                    return json.loads(f.read())
            except (OSError, ValueError):
                # We'll overwrite the corrupted cache file later
                pass

    def _store_info_to_cache(self, image_path, info):
        cache_file = self._cache_file(image_path)
        try:
            json_string = json.dumps(info, indent=4) + '\n'
        except (TypeError, ValueError) as e:
            raise RuntimeError(f'Unable to write cache {cache_file}: {e}')

        try:
            fs.mkdir(fs.dirname(cache_file))
            with open(cache_file, 'w') as f:
                f.write(json_string)
        except (OSError, TypeError, ValueError) as e:
            msg = e.strerror if getattr(e, 'strerror', None) else e
            raise RuntimeError(f'Unable to write cache {cache_file}: {msg}')

    def _cache_file(self, image_path):
        # If image is in our cache_dir, the image's file name makes it
        # unique. This is usually the case when we're uploading screenshots. If
        # image is not in our cache_dir, use the absolute path as a unique
        # identifier.
        if fs.dirname(image_path) == self._cache_dir:
            image_path = os.path.basename(image_path)
        else:
            image_path = os.path.abspath(image_path)
        # Max file name length is ususally 255 bytes
        filename = fs.sanitize_filename(image_path[-200:]) + f'.{self.name}.json'
        return os.path.join(self._cache_dir, filename)



from base64 import b64decode

from .. import http
from . import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


class SrrDbApi(base.SceneDbApiBase):
    name = 'srrdb'
    label = 'srrDB'

    _url_base = b64decode('d3d3LnNycmRiLmNvbQ==').decode('ascii')
    _search_url = f'https://{_url_base}/api/search'
    _details_url = f'https://{_url_base}/api/details'

    async def _search(self, keywords, group):
        if group:
            keywords = list(keywords)
            keywords.append(f'group:{group}')
        keywords_path = '/'.join((kw.lower() for kw in keywords))
        search_url = f'{self._search_url}/{keywords_path}'
        _log.debug('Scene search URL: %r', search_url)
        response = (await http.get(search_url, cache=True)).json()
        results = response.get('results', [])
        return (r['release'] for r in results)

    async def release_files(self, release_name):
        """
        Map file names to dictionaries with the keys ``release_name``,
        ``file_name``, ``size`` and ``crc``

        :param str release_name: Exact name of the release

        :raise RequestError: if request fails
        """
        details_url = f'{self._details_url}/{release_name}'
        _log.debug('Scene details URL: %r', details_url)
        response = (await http.get(details_url, cache=True)).json()
        if not response:
            return {}
        else:
            files = response.get('archived-files', ())
            release_name = response.get('name', '')
            return {
                f['name']: {
                    'release_name': release_name,
                    'file_name': f['name'],
                    'size': f['size'],
                    'crc': f['crc'],
                }
                for f in sorted(files, key=lambda f: f['name'].casefold())
            }

    def release_url(self, release_name):
        """Return website URL for particular release"""
        return f'https://{self._url_base}/release/details/{release_name}'



"""
BbTrackerConfig class
"""

import base64

from ..base import TrackerConfigBase


class BbTrackerConfig(TrackerConfigBase):
    defaults = {
        'base_url'   : base64.b64decode('aHR0cHM6Ly9iYWNvbmJpdHMub3Jn').decode('ascii'),
        'username'   : '',
        'password'   : '',
        'source'     : 'bB',
        'image_host' : 'imgbox',
        'exclude'    : [
            r'\.(?i:nfo|txt|jpg|jpeg|png|sfv|md5)$',
            r'/(?i:sample|extra|bonus|feature)',
        ],
    }



"""
Abstract base class for scene release databases
"""

import abc

import logging  # isort:skip
_log = logging.getLogger(__name__)


class SceneDbApiBase(abc.ABC):
    """Base class for scene release database APIs"""

    @property
    @abc.abstractmethod
    def name(self):
        """Unique name of the scene release database"""

    @property
    @abc.abstractmethod
    def label(self):
        """User-facing name of the scene release database"""

    async def search(self, query, only_existing_releases=True):
        """
        Search for scene release

        :param SceneQuery query: Search query
        :param bool only_existing_releases: See
            :meth:`~.scene.find.SceneQuery.search`

        :return: :class:`list` of release names as :class:`str`

        :raise RequestError: if the search request fails
        """
        return await query.search(
            self._search,
            only_existing_releases=only_existing_releases,
        )

    @abc.abstractmethod
    async def _search(self, keywords, group=None):
        pass

    @abc.abstractmethod
    async def release_files(self, release_name):
        """
        Map file names to file objects

        What a file object is depends on the subclass implementation.

        :param str release_name: Exact name of the release

        :raise RequestError: if request fails
        """



"""
Background workers to keep the UI responsive
"""

import asyncio
import enum
import functools
import multiprocessing

from .. import errors

import logging  # isort:skip
_log = logging.getLogger(__name__)


class MsgType(enum.Enum):
    """
    Enum that specifies the type of an IPC message (info, error, etc)
    """
    init = 'init'
    info = 'info'
    error = 'error'
    result = 'result'
    terminate = 'terminate'


class DaemonProcess:
    """
    :class:`multiprocessing.Process` abstraction with IPC

    Intended to offload heavy work (e.g. torrent creation) onto a different
    process. (Threads can still make the UI unresponsive because of the
    `GIL <https://wiki.python.org/moin/GlobalInterpreterLock>`_.)
    """

    def __init__(self, target, name=None, args=(), kwargs={},
                 init_callback=None, info_callback=None, error_callback=None,
                 result_callback=None, finished_callback=None):
        self._target = target
        self._name = name
        self._target_args = args
        self._target_kwargs = kwargs
        self._init_callback = init_callback
        self._info_callback = info_callback
        self._error_callback = error_callback
        self._result_callback = result_callback
        self._finished_callback = finished_callback

        self._ctx = multiprocessing.get_context('spawn')
        self._write_queue = self._ctx.Queue()
        self._read_queue = self._ctx.Queue()
        self._process = None
        self._read_queue_reader_task = None
        self._exception = None
        self._loop = asyncio.get_event_loop()

    def start(self):
        """Start the process"""
        _log.debug('Starting process: %r', self._target)
        target_wrapped = functools.partial(
            _target_process_wrapper,
            self._target,
            self._read_queue,
            self._write_queue,
        )
        self._process = self._ctx.Process(
            name=self._name,
            target=target_wrapped,
            args=self._target_args,
            kwargs=self._target_kwargs,
        )
        self._process.start()
        self._read_queue_reader_task = self._loop.create_task(self._read_queue_reader())
        self._read_queue_reader_task.add_done_callback(self._handle_read_queue_reader_task_done)

    async def _read_queue_reader(self):
        while True:
            typ, msg = await self._loop.run_in_executor(None, self._read_queue.get)
            if typ is MsgType.init:
                if self._init_callback:
                    self._init_callback(msg)
            elif typ is MsgType.info:
                if self._info_callback:
                    self._info_callback(msg)
            elif typ is MsgType.error:
                if isinstance(msg, tuple):
                    exception, traceback = msg
                    raise errors.SubprocessError(exception, traceback)
                elif isinstance(msg, BaseException):
                    raise msg
                else:
                    if self._error_callback:
                        self._error_callback(msg)
            elif typ is MsgType.result:
                if self._result_callback:
                    self._result_callback(msg)
                break
            elif typ is MsgType.terminate:
                break
            else:
                raise RuntimeError(f'Unknown message type: {typ!r}')

    def _handle_read_queue_reader_task_done(self, task):
        # Catch exceptions from _read_queue_reader()
        try:
            task.result()
        except BaseException as e:
            _log.debug('Caught _read_queue_reader() exception: %r', e)
            self._exception = e
            if self._error_callback:
                self._error_callback(e)
            else:
                raise self._exception
        finally:
            if self._finished_callback:
                try:
                    self._finished_callback()
                except BaseException as e:
                    self._exception = e
                    if self._error_callback:
                        self._error_callback(e)
                    raise self._exception

    def stop(self):
        """Stop the process"""
        if self.is_alive:
            self._write_queue.put((MsgType.terminate, None))

    @property
    def is_alive(self):
        """Whether :meth:`start` was called and the process has not finished yet"""
        if self._process:
            return self._process.is_alive()
        else:
            return False

    @property
    def exception(self):
        """Exception from `target` or callback, `None` if no exception was raised"""
        task = self._read_queue_reader_task
        if task and task.done() and task.exception():
            return task.exception()
        else:
            return self._exception

    async def join(self):
        """Block asynchronously until the process exits"""
        if self._process:
            _log.debug('Joining process: %r', self._process)
            await self._loop.run_in_executor(None, self._process.join)
            self._process = None

        if self._read_queue_reader_task:
            if not self._read_queue_reader_task.done():
                # Unblock self._read_queue()
                self._read_queue.put((MsgType.terminate, None))
                await self._read_queue_reader_task
            exc = self.exception
            if exc:
                raise exc


def _target_process_wrapper(target, write_queue, read_queue, *args, **kwargs):
    try:
        target(write_queue, read_queue, *args, **kwargs)
    except BaseException as e:
        # Because the traceback is not picklable, preserve it as a string before
        # sending it over the Queue
        import traceback
        traceback = ''.join(traceback.format_exception(type(e), e, e.__traceback__))
        write_queue.put((MsgType.error, (e, traceback)))
    finally:
        write_queue.put((MsgType.terminate, None))



"""
Classes and functions that are used by all :class:`~.base.WebDbApiBase`
subclasses
"""

import re

from .. import release
from ..types import ReleaseType

import logging  # isort:skip
_log = logging.getLogger(__name__)


class Query:
    """
    Search query for databases like IMDb

    :param str title: Name of the movie or TV series
    :param str type: :class:`~.types.ReleaseType` enum or one of its value names
    :param year: Year of release
    :type year: str or int

    :raise ValueError: if an invalid argument is passed
    """

    @staticmethod
    def _normalize_title(title):
        return ' '.join(title.casefold().strip().split())

    def __init__(self, title, year=None, type=ReleaseType.unknown):
        self._type = ReleaseType(type)
        self._title = str(title)
        self._title_normalized = self._normalize_title(self._title)
        self._year = None if year is None else str(year)
        if self._year is not None and not 1800 < int(self._year) < 2100:
            raise ValueError(f'Invalid year: {self._year}')

    @property
    def type(self):
        """:class:`~.types.ReleaseType` value"""
        return self._type

    @property
    def title(self):
        """Name of the movie or TV series"""
        return self._title

    @property
    def title_normalized(self):
        """Same as :attr:`title` but in stripped lower case with deduplicated spaces"""
        return self._title_normalized

    @property
    def year(self):
        """Year of release"""
        return self._year

    _types = {
        ReleaseType.movie: ('movie', 'film'),
        ReleaseType.season: ('season', 'series', 'tv', 'show', 'tvshow'),
        ReleaseType.episode: ('episode',),
    }
    _known_types = tuple(v for types in _types.values() for v in types)
    _kw_regex = {
        'year': r'year:(\d{4})',
        'type': rf'type:({"|".join(_known_types)})',
    }

    @classmethod
    def from_string(cls, query):
        """
        Create instance from string

        The returned :class:`Query` is case-insensitive and has any superfluous
        whitespace removed.

        Keyword arguments are extracted by looking for ``"year:YEAR"`` and
        ``"type:TYPE"`` in `query` where ``YEAR`` is a four-digit number and
        ``TYPE`` is something like "movie", "film", "tv", etc.
        """
        # Normalize query:
        #   - Case-insensitive
        #   - Remove leading/trailing white space
        #   - Deduplicate white space
        query = ' '.join(str(query).strip().split())

        def get_kwarg(string):
            for kw, regex in cls._kw_regex.items():
                match = re.search(f'^{regex}$', part)
                if match:
                    value = match.group(1)
                    if kw == 'type' and value in cls._types[ReleaseType.movie]:
                        return 'type', ReleaseType.movie
                    elif kw == 'type' and value in cls._types[ReleaseType.season]:
                        return 'type', ReleaseType.season
                    elif kw == 'type' and value in cls._types[ReleaseType.episode]:
                        return 'type', ReleaseType.episode
                    elif kw == 'year':
                        return 'year', value
            return None, None

        # Extract "key:value" pairs (e.g. "year:2015")
        title = []
        kwargs = {}
        for part in query.split():
            kw, value = get_kwarg(part)
            if (kw, value) != (None, None):
                kwargs[kw] = value
            else:
                title.append(part)

        return cls(title=' '.join(title), **kwargs)

    @classmethod
    def from_path(cls, path):
        """
        Create instance from file or directory name

        `path` is passed to :class:`~.release.ReleaseInfo` to get the
        arguments for instantiation.
        """
        info = release.ReleaseInfo(path)
        kwargs = {'title': info['title']}
        if info.get('year'):
            kwargs['year'] = info['year']
        if info.get('type'):
            kwargs['type'] = info['type']
        return cls(**kwargs)

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (
                self.title_normalized == other.title_normalized
                and self.year == other.year
                and self.type is other.type
            )
        else:
            return NotImplemented

    def __str__(self):
        parts = [self.title]
        if self.year:
            parts.append(f'year:{self.year}')
        if self.type is not ReleaseType.unknown:
            parts.append(f'type:{self.type}')
        return ' '.join(parts)

    def __repr__(self):
        kwargs = ', '.join(
            f'{k}={v!r}'
            for k, v in (('title', self.title),
                         ('year', self.year),
                         ('type', self.type))
            if v not in (None, ReleaseType.unknown)
        )
        return f'{type(self).__name__}({kwargs})'


class SearchResult:
    """
    Information about a search result

    Keyword arguments are available as attributes.

    `cast`, `countries`, `director`, `keywords`, `summary`, `title_english` and
    `title_original` may be coroutine functions that return the expected value.

    :param str director: Name of director
    :param cast: Short list of actor names
    :type cast: sequence of str
    :param str countries: List of country names of origin
    :param str id: ID for the relevant DB
    :param keywords: Short list of keywords, e.g. `["horror", "commedy"]`
    :type keywords: sequence of str
    :param summary: Short text that describes the movie or series
    :param str title: Title of the movie or series
    :param str title_english: English title of the movie or series
    :param str title_original: Original title of the movie or series
    :param str type: :class:`~.types.ReleaseType` value
    :param str url: Web page of the search result
    :param str year: Release year; for series this should be the year of the
        first airing of the first episode of the first season
    """
    def __init__(self, *, id, type, url, year, cast=(), countries=(), director='',
                 keywords=(), summary='', title, title_english='',
                 title_original=''):
        self._info = {
            'id' : id,
            'title' : str(title),
            'type' : ReleaseType(type),
            'url' : str(url),
            'year' : str(year),
            # These may be coroutine functions
            'cast' : cast,
            'countries' : countries,
            'director' : director,
            'keywords' : keywords,
            'summary' : summary,
            'title_english' : title_english,
            'title_original' : title_original,
        }

    def __getattr__(self, name):
        try:
            return self._info[name]
        except KeyError:
            raise AttributeError(name)

    def __repr__(self):
        kwargs = ', '.join(f'{k}={v!r}' for k, v in self._info.items())
        return f'{type(self).__name__}({kwargs})'


class Person(str):
    """:class:`str` subclass with an `url` attribute"""

    def __new__(cls, string, url=''):
        obj = super().__new__(cls, string)
        obj.url = url
        return obj



from ....utils import cached_property
from .. import widgets
from . import JobWidgetBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class ChoiceJobWidget(JobWidgetBase):
    def setup(self):
        self._radiolist = widgets.RadioList(
            choices=self.job.choices,
            focused=self.job.focused,
            on_accepted=self._handle_accepted,
        )
        self.job.signal.register('dialog_updated', self._handle_dialog_updated)

    def _handle_dialog_updated(self, job):
        self._radiolist.choices = job.choices
        self._radiolist.focused_index = job.focused_index
        self.invalidate()

    def _handle_accepted(self, choice):
        # This also finishes self.job
        self.job.choice = choice

    @cached_property
    def runtime_widget(self):
        return self._radiolist


class TextFieldJobWidget(JobWidgetBase):
    def setup(self):
        self._input_field = widgets.InputField(
            text=self.job.text,
            read_only=self.job.read_only,
            style='class:dialog.text',
            on_accepted=self._handle_accepted,
        )
        self.job.signal.register('dialog_updated', self._handle_dialog_updated)

    def _handle_dialog_updated(self, job):
        self._input_field.read_only = job.read_only
        self._input_field.text = job.text
        self.invalidate()

    def _handle_accepted(self, buffer):
        self.job.send(buffer.text)

    @cached_property
    def runtime_widget(self):
        return self._input_field



"""
Main UI widget and job manager
"""

import asyncio
import collections
import types

from prompt_toolkit.application import Application
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import Layout
from prompt_toolkit.layout.containers import HSplit, Window, to_container

from . import jobwidgets, style

import logging  # isort:skip
_log = logging.getLogger(__name__)


class UI:
    def __init__(self):
        # Map JobBase.name to SimpleNamespace with attributes:
        #   job       - JobBase instance
        #   widget    - JobWidgetBase instance
        #   container - Container instance
        self._jobs = collections.defaultdict(lambda: types.SimpleNamespace())
        self._app = self._make_app()
        self._app_terminated = False
        self._exception = None
        self._loop = asyncio.get_event_loop()
        self._loop.set_exception_handler(self._handle_exception)

    def _handle_exception(self, loop, context):
        exception = context.get('exception')
        if exception:
            _log.debug('Caught unhandled exception: %r', exception)
            if not self._exception:
                self._exception = exception
            self._exit()

    def _make_app(self):
        self._jobs_container = HSplit(
            # FIXME: Layout does not accept an empty list of children. We add an
            #        empty Window that doesn't display anything that gets
            #        removed automatically when we rebuild
            #        self._jobs_container.children.
            #        https://github.com/prompt-toolkit/python-prompt-toolkit/issues/1257
            children=[Window()],
            style='class:default',
        )
        self._layout = Layout(self._jobs_container)

        kb = KeyBindings()

        @kb.add('escape')
        @kb.add('c-g')
        @kb.add('c-q')
        @kb.add('c-c')
        def _(event, self=self):
            if self._app.is_running:
                self._exit()

        app = Application(
            layout=self._layout,
            key_bindings=kb,
            style=style.style,
            full_screen=False,
            erase_when_done=False,
            mouse_support=False,
        )
        # Make escape key work
        app.timeoutlen = 0.1
        app.ttimeoutlen = 0.1
        return app

    def add_jobs(self, *jobs):
        """Add :class:`~.jobs.base.JobBase` instances"""
        for job in jobs:
            if job.name in self._jobs:
                raise RuntimeError(f'Job was already added: {job.name}')
            else:
                self._jobs[job.name].job = job
                self._jobs[job.name].widget = jobwidgets.JobWidget(job, self._app)
                self._jobs[job.name].container = to_container(self._jobs[job.name].widget)

                # Terminate application if all jobs finished
                job.signal.register('finished', self._exit_if_all_jobs_finished)

                # Terminate application if any job finished with non-zero exit code
                job.signal.register('finished', self._exit_if_job_failed)

                if self._jobs[job.name].widget.is_interactive:
                    # Display next interactive job when currently focused job finishes
                    job.signal.register('finished', self._update_jobs_container)

        self._update_jobs_container()

    def _update_jobs_container(self, *_):
        job_containers = []

        # Ensure enabled jobs are started
        for jobinfo in self._enabled_jobs:
            if not jobinfo.job.is_started and jobinfo.job.autostart:
                jobinfo.job.start()

        # List interactive jobs first
        for jobinfo in self._enabled_jobs:
            if jobinfo.widget.is_interactive:
                job_containers.append(jobinfo.container)
                # Focus the first unfinished job
                if not jobinfo.job.is_finished:
                    _log.debug('Active job: %r', jobinfo.job.name)
                    try:
                        self._layout.focus(jobinfo.container)
                    except ValueError:
                        pass
                    # Don't display more than one unfinished interactive job
                    # unless any job has errors, in which case we are
                    # terminating the application and display all jobs.
                    if not any(jobinfo.job.errors for jobinfo in self._jobs.values()):
                        break

        # Add non-interactive jobs below interactive jobs so the interactive
        # widgets don't change position when non-interactive widgets change
        # size.
        for jobinfo in self._enabled_jobs:
            if not jobinfo.widget.is_interactive:
                job_containers.append(jobinfo.container)

        # Replace visible containers
        self._jobs_container.children[:] = job_containers

    @property
    def _enabled_jobs(self):
        return tuple(jobinfo for jobinfo in self._jobs.values()
                     if jobinfo.job.is_enabled)

    def run(self, jobs):
        """
        Block while running `jobs`

        :param jobs: Iterable of :class:`~.jobs.base.JobBase` instances

        :raise: Any exception that occured while running jobs

        :return: :attr:`~.JobBase.exit_code` from the first failed job or 0 for
            success
        """
        self.add_jobs(*jobs)

        # Block until _exit() is called
        self._app.run(set_exception_handler=False)

        exception = self._get_exception()
        if exception:
            _log.debug('Application exception: %r', exception)
            raise exception
        else:
            # First non-zero exit_code is the application exit_code
            for jobinfo in self._enabled_jobs:
                _log.debug('Checking exit_code of %r: %r', jobinfo.job.name, jobinfo.job.exit_code)
                if jobinfo.job.exit_code != 0:
                    return jobinfo.job.exit_code
            return 0

    def _exit_if_all_jobs_finished(self, *_):
        if all(jobinfo.job.is_finished for jobinfo in self._enabled_jobs):
            _log.debug('All jobs finished')
            self._exit()

    def _exit_if_job_failed(self, job):
        if job.is_finished and job.exit_code != 0:
            _log.debug('Terminating application because of failed job: %r', job.name)
            self._exit()

    def _exit(self):
        if not self._app_terminated:
            if not self._app.is_running and not self._app.is_done:
                self._loop.call_soon(self._exit)
            else:
                def when_jobs_terminated():
                    self._app.exit()
                    self._update_jobs_container()

                self._app_terminated = True
                self._app.create_background_task(
                    self._terminate_jobs(callback=when_jobs_terminated),
                )

    async def _terminate_jobs(self, callback=None):
        _log.debug('Waiting for jobs before exiting')
        self._finish_jobs()
        for jobinfo in self._enabled_jobs:
            if jobinfo.job.is_started and not jobinfo.job.is_finished:
                _log.debug('Waiting for %r', jobinfo.job.name)
                await jobinfo.job.wait()
                _log.debug('Waiting for %r', jobinfo.job.name)
        if callback:
            callback()

    def _finish_jobs(self):
        for jobinfo in self._enabled_jobs:
            if not jobinfo.job.is_finished:
                _log.debug('Finishing %s', jobinfo.job.name)
                jobinfo.job.finish()

    def _get_exception(self):
        if self._exception:
            # Exception from _handle_exception()
            return self._exception
        else:
            # First exception from jobs
            for jobinfo in self._enabled_jobs:
                if jobinfo.job.raised:
                    _log.debug('Exception from %s: %r', jobinfo.job.name, jobinfo.job.raised)
                    return jobinfo.job.raised



"""
Exception classes

Abstraction layers should raise one of these exceptions if an error message
should be displayed to the user. If the programmer made a mistake, any
appropriate builtin exception (e.g. :class:`ValueError`, :class:`TypeError`) or
:class:`RuntimeError` should be raised.

For example, :func:`upsies.utils.http.get` always raises :class:`RequestError`,
regardless of which library is used or what went wrong, except when something
like caching fails, which is most likely due to a bug.
"""

class UpsiesError(Exception):
    """Base class for all exceptions raised by upsies"""

    def __eq__(self, other):
        if isinstance(other, type(self)) and str(other) == str(self):
            return True
        else:
            return NotImplemented


class CancelledError(UpsiesError):
    """User cancelled an operation"""


class ConfigError(UpsiesError):
    """Error while reading/writing config file or setting config file option"""


class DependencyError(UpsiesError):
    """Some external tool is missing (e.g. ``mediainfo``)"""


class ContentError(UpsiesError):
    """
    Something is wrong with user-provided content, e.g. no video files in the
    given directory or no permission to read
    """

class ProcessError(UpsiesError):
    """Executing subprocess failed"""


class RequestError(UpsiesError):
    """Network request failed"""
    def __init__(self, msg, headers={}, status_code=None):
        super().__init__(msg)
        self._headers = headers
        self._status_code = status_code

    @property
    def headers(self):
        """HTTP headers from server response or empty `dict`"""
        return self._headers

    @property
    def status_code(self):
        """HTTP status code (e.g. 404) or `None`"""
        return self._status_code


class ScreenshotError(UpsiesError):
    """Screenshot creation failed"""


class ImageResizeError(UpsiesError):
    """Image resizing failed"""


class TorrentError(UpsiesError):
    """Torrent file creation failed"""


def SubprocessError(exception, original_traceback):
    """Attach `original_traceback` to `exception`"""
    exception.original_traceback = f'Subprocess traceback:\n{original_traceback.strip()}'
    return exception


class SceneError(UpsiesError):
    """Base class for scene-related errors"""

class SceneRenamedError(SceneError):
    """Renamed scene release"""
    def __init__(self, *, original_name, existing_name):
        super().__init__(f'Release name should be: {original_name}')
        self._original_name = original_name
        self._existing_name = existing_name

    @property
    def original_name(self):
        """What the release name should be"""
        return self._original_name

    @property
    def existing_name(self):
        """What the release name is"""
        return self._existing_name

    @property
    def url(self):
        """URL to scene database for this release"""
        from .utils import scene
        db = scene.scenedb('srrdb')
        if hasattr(db, 'release_url'):
            return db.release_url(self._original_name)

class SceneFileSizeError(SceneError):
    """Scene release file size differs from original release"""
    def __init__(self, filename, *, original_size, existing_size):
        super().__init__(f'{filename} should be {original_size} bytes, not {existing_size}')
        self._filename = filename
        self._original_size = original_size
        self._existing_size = existing_size

    @property
    def filename(self):
        """Name of the file in the scene release"""
        return self._filename

    @property
    def original_size(self):
        """Size of the file in the scene release"""
        return self._original_size

    @property
    def existing_size(self):
        """Size of the file in the local file system"""
        return self._existing_size

class SceneMissingInfoError(UpsiesError):
    """Missing information about a file from a scene release"""
    def __init__(self, file_name):
        super().__init__(f'Missing information: {file_name}')



"""
Dummy tracker for testing and debugging
"""

import asyncio
import os

from .. import errors, jobs
from ..utils import cached_property
from ..utils.types import ReleaseType
from . import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


class DummyTrackerConfig(base.TrackerConfigBase):
    defaults = {
        'base_url'   : 'http://localhost',
        'username'   : '',
        'password'   : '',
        'exclude'    : [],
        'source'     : 'DMY',
        'image_host' : 'dummy',
    }


class DummyTrackerJobs(base.TrackerJobsBase):
    @cached_property
    def jobs_before_upload(self):
        return (
            self.create_torrent_job,
            self.screenshots_job,
            self.upload_screenshots_job,
            self.mediainfo_job,
            self.imdb_job,
            self.tmdb_job,
            self.release_name_job,
            self.category_job,
            self.scene_check_job,
        )

    @cached_property
    def category_job(self):
        if not self.cli_args.skip_category:

            def select_release_category(release_name):
                _log.debug('Setting type: %r', release_name.type)
                if release_name.type:
                    self.category_job.focused = release_name.type

            self.release_name_job.signal.register('release_name', select_release_category)

            choices = [(str(t).capitalize(), t) for t in ReleaseType if t]
            return jobs.dialog.ChoiceJob(
                name='category',
                label='Category',
                choices=choices,
                **self.common_job_args,
            )


class DummyTracker(base.TrackerBase):
    name = 'dummy'
    label = 'DuMmY'

    argument_definitions = {
        ('--skip-category', '-C'): {
            'help': 'Do not ask for category',
            'action': 'store_true',
        },
        ('--screenshots', '-s'): {
            'help': 'How many screenshots to make',
            'type': int,
            'default': 3,
        },
        ('--delay', '-d'): {
            'help': 'Number of seconds login, upload and logout take each',
            'type': float,
            'default': 1.0,
        },
    }

    TrackerJobs = DummyTrackerJobs
    TrackerConfig = DummyTrackerConfig

    async def login(self):
        _log.debug('%s: Logging in with %r', self.name, self.config)
        await asyncio.sleep(self.cli_args.delay)
        self._is_logged_in = True

    async def logout(self):
        _log.debug('%s: Logging out', self.name)
        await asyncio.sleep(self.cli_args.delay)
        self._is_logged_in = False

    @property
    def is_logged_in(self):
        return getattr(self, '_is_logged_in', False)

    async def get_announce_url(self):
        _log.debug('%s: Getting announce URL', self.name)
        await asyncio.sleep(self.cli_args.delay)
        return 'http://localhost:123/f1dd15718/announce'

    async def upload(self, tracker_jobs):
        if tracker_jobs.create_torrent_job.output:
            torrent_file = tracker_jobs.create_torrent_job.output[0]
        else:
            raise errors.RequestError('Torrent file was not created.')
        _log.debug('%s: Uploading %s', self.name, torrent_file)
        await asyncio.sleep(self.cli_args.delay)
        return f'http://localhost/{os.path.basename(torrent_file)}'



from ....utils import cached_property
from .. import widgets
from . import JobWidgetBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class CreateTorrentJobWidget(JobWidgetBase):
    def setup(self):
        self._progress = widgets.ProgressBar()
        self.job.signal.register('progress_update', self.handle_progress_update)
        self.job.signal.register('finished', lambda _: self.invalidate())

    def handle_progress_update(self, percent_done):
        self._progress.percent = percent_done
        self.invalidate()

    @cached_property
    def runtime_widget(self):
        return self._progress


class AddTorrentJobWidget(JobWidgetBase):
    def setup(self):
        pass

    @cached_property
    def runtime_widget(self):
        return None


class CopyTorrentJobWidget(JobWidgetBase):
    def setup(self):
        pass

    @cached_property
    def runtime_widget(self):
        return None



"""
NBL API
"""

import base64
import re
import urllib

from .. import errors, jobs
from ..utils import cached_property, html, http, release
from . import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


class NblTrackerConfig(base.TrackerConfigBase):
    defaults = {
        'base_url'   : base64.b64decode('aHR0cHM6Ly9uZWJ1bGFuY2UuaW8=').decode('ascii'),
        'username'   : '',
        'password'   : '',
        'exclude'    : [],
        'source'     : 'NBL',
        'image_host' : '',
    }


class NblTrackerJobs(base.TrackerJobsBase):
    @cached_property
    def jobs_before_upload(self):
        return (
            self.create_torrent_job,
            self.mediainfo_job,
            self.tvmaze_job,
            self.category_job,
        )

    @cached_property
    def category_job(self):
        # Season or Episode
        release_info = release.ReleaseInfo(self.content_path)
        if release_info['type'] == release.ReleaseType.episode:
            guessed = 'Episode'
        else:
            guessed = 'Season'
        return jobs.dialog.ChoiceJob(
            name='category',
            label='Category',
            choices=('Season', 'Episode'),
            focused=guessed,
            **self.common_job_args,
        )


class NblTracker(base.TrackerBase):
    name = 'nbl'
    label = 'NBL'

    argument_definitions = {
        ('--ignore-dupes', '-D'): {
            'help': 'Force submission even if the tracker reports duplicates',
            'action': 'store_true',
        },
    }

    TrackerConfig = NblTrackerConfig
    TrackerJobs = NblTrackerJobs

    _url_path = {
        'login': '/login.php',
        'upload': '/upload.php',
        'torrent': '/torrents.php',
    }

    async def login(self):
        if not self.is_logged_in:
            if not self.config.get('username'):
                raise errors.RequestError('Login failed: No username configured')
            elif not self.config.get('password'):
                raise errors.RequestError('Login failed: No password configured')

            _log.debug('%s: Logging in as %r', self.name, self.config['username'])
            login_url = urllib.parse.urljoin(
                self.config['base_url'],
                self._url_path['login'],
            )
            response = await http.post(
                url=login_url,
                user_agent=True,
                data={
                    'username': self.config['username'],
                    'password': self.config['password'],
                    'twofa': '',
                    'login': 'Login',
                },
            )
            try:
                doc = html.parse(response)
                self._report_login_error(doc)
                self._store_auth_key(doc)
                self._store_logout_url(doc)
            except Exception:
                html.dump(str(response), 'login.html')
                raise

    def _report_login_error(self, doc):
        def error_tag(tag):
            class_ = tag.get('class', ())
            return (
                'warning' in class_
                and 'hidden' not in class_
                and 'noscript' not in (p.name for p in tag.parents)
            )

        error = doc.find(error_tag)
        if error:
            msg = ' '.join(error.stripped_strings).strip()
            if msg:
                raise errors.RequestError(f'Login failed: {msg}')

    def _store_auth_key(self, doc):
        auth_input = doc.find('input', {'name': 'auth'})
        if not auth_input:
            raise RuntimeError('Failed to find input tag named "auth"')
        else:
            self._auth_key = auth_input['value']
            _log.debug('%s: Auth key: %s', self.name, self._auth_key)

    def _store_logout_url(self, doc):
        logout_url = doc.find('a', text=re.compile(r'(?i:Logout)'))
        if not logout_url or not logout_url.get('href'):
            raise RuntimeError('Failed to find logout URL')
        else:
            _log.debug('%s: Logged in as %r', self.name, self.config['username'])
            self._logout_url = urllib.parse.urljoin(
                self.config['base_url'],
                logout_url['href'],
            )
            _log.debug('%s: Logout URL: %s', self.name, self._logout_url)

    @property
    def is_logged_in(self):
        return all(hasattr(self, attr)
                   for attr in ('_logout_url', '_auth_key'))

    async def logout(self):
        if hasattr(self, '_auth_key'):
            delattr(self, '_auth_key')
        if hasattr(self, '_logout_url'):
            logout_url = self._logout_url
            delattr(self, '_logout_url')
            _log.debug('%s: Logging out: %r', self.name, logout_url)
            await http.get(logout_url, user_agent=True)

    async def get_announce_url(self):
        url = urllib.parse.urljoin(
            self.config['base_url'],
            self._url_path['upload'],
        )
        _log.debug('%s: Getting announce URL from %s', self.name, url)
        response = await http.get(url, cache=False, user_agent=True)
        doc = html.parse(response)
        announce_url_tag = doc.find('input', value=re.compile(r'^https?://.*/announce$'))
        if announce_url_tag:
            _log.debug('%s: Announce URL: %s', self.name, announce_url_tag['value'])
            return announce_url_tag['value']

    async def upload(self, tracker_jobs):
        if not self.is_logged_in:
            raise RuntimeError('upload() called before login()')

        metadata = self._get_metadata_from_jobs(tracker_jobs)
        _log.debug('%s: Torrent: %r', self.name, metadata['torrent_filepath'])
        _log.debug('%s: TVmaze ID: %r', self.name, metadata['tvmaze_id'])
        _log.debug('%s: Mediainfo: %r', self.name, metadata['mediainfo'][:100] + '...')
        _log.debug('%s: Category: %r', self.name, metadata['category'])

        post_data = {
            'auth': self._auth_key,
            'title': '',
            'tvmazeid': metadata['tvmaze_id'],
            'media': metadata['mediainfo'],
            'desc': metadata['mediainfo'],
            'mediaclean': f'[mediainfo]{metadata["mediainfo"]}[/mediainfo]',
            'submit': 'true',
            'MAX_FILE_SIZE': '1048576',
            'category': metadata['category'],
            'genre_tags': '',
            'tags': '',
            'image': '',
            'fontfont': '-1',
            'fontsize': '-1',
        }
        if getattr(self.cli_args, 'ignore_dupes', None):
            post_data['ignoredupes'] = '1'

        upload_url = urllib.parse.urljoin(
            self.config['base_url'],
            self._url_path['upload'],
        )

        response = await http.post(
            url=upload_url,
            user_agent=True,
            allow_redirects=False,
            files={'file_input': (metadata['torrent_filepath'], 'application/x-bittorrent')},
            data=post_data,
        )

        # Upload response should redirect to torrent page via "Location" header
        torrent_page_url = urllib.parse.urljoin(
            self.config['base_url'],
            response.headers.get('Location', ''),
        )
        if urllib.parse.urlparse(torrent_page_url).path == self._url_path['torrent']:
            return str(torrent_page_url)
        else:
            _log.debug('Unexpected torrent page URL: %r', torrent_page_url)
            doc = html.parse(response)
            # Try to find error message
            error = doc.find(id='messagebar')
            if error and error.string:
                raise errors.RequestError(f'Upload failed: {error.string}')
            else:
                html.dump(str(response), 'upload.html')
                raise RuntimeError('Failed to find error message. See upload.html for more information.')

    def _get_metadata_from_jobs(self, tracker_jobs):

        def translate_category(category):
            if category.lower() == 'episode':
                return '1'
            elif category.lower() == 'season':
                return '3'
            else:
                raise errors.RequestError(f'Unsupported type: {category}')

        return {
            'torrent_filepath': tracker_jobs.create_torrent_job.output[0],
            'tvmaze_id': tracker_jobs.tvmaze_job.output[0],
            'mediainfo': tracker_jobs.mediainfo_job.output[0],
            'category': translate_category(tracker_jobs.category_job.output[0]),
        }



"""
Generate uniform release name
"""

import asyncio

from ..utils import fs, release
from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class ReleaseNameJob(JobBase):
    """
    Create user-confirmed release name from file or directory path

    This job adds the following signals to the :attr:`~.JobBase.signal`
    attribute:

        ``release_name_updated``
            Emitted after :meth:`~.ReleaseName.fetch_info` updated the release
            name. Registered callbacks get a :class:`~.release.ReleaseName`
            instance as a positional argument.

        ``release_name``
            Emitted after the user approved a release name. Registered callbacks
            get a :class:`~.release.ReleaseName` instance as a positional
            argument.
    """

    name = 'release-name'
    label = 'Release Name'

    @property
    def cache_id(self):
        """Final segment of `content_path`"""
        return fs.basename(self._content_path)

    @property
    def release_name(self):
        """:class:`~.utils.release.ReleaseName` instance"""
        return self._release_name

    def initialize(self, *, content_path):
        """
        Set internal state

        :param content_path: Path to release file or directory
        """
        self._content_path = content_path
        self._release_name = release.ReleaseName(content_path)
        self.signal.add('release_name_updated')
        self.signal.add('release_name')

    def release_name_selected(self, release_name):
        """
        :meth:`Send <upsies.jobs.base.JobBase.send>` release name and :meth:`finish
        <upsies.jobs.base.JobBase.finish>`

        This method must be called by the UI when the user accepts the release
        name.

        :param str release_name: Release name as accepted by the user
        """
        # Parse string to get ReleaseName instance for the "release_name" signal
        self._release_name = release.ReleaseName(release_name)
        self.signal.emit('release_name', self.release_name)
        # Send original string as we received it
        self.send(release_name)
        self.finish()

    def fetch_info(self, *args, **kwargs):
        """
        Synchronous wrapper around :meth:`.ReleaseName.fetch_info` that emits the
        ``release_name_updated`` signal

        All arguments are passed on to :meth:`.ReleaseName.fetch_info`.
        """
        task = asyncio.ensure_future(
            self.release_name.fetch_info(*args, **kwargs),
        )

        def fetch_info_done(task):
            exc = task.exception()
            if exc:
                self.exception(exc)
            else:
                self.signal.emit('release_name_updated', self.release_name)

        task.add_done_callback(fetch_info_done)



"""
Custom job that provide the result of a coroutine
"""

import asyncio

from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class CustomJob(JobBase):
    """Wrapper around coroutine function or awaitable"""

    @property
    def name(self):
        return self._name

    @property
    def label(self):
        return self._label

    def initialize(self, *, name, label, worker):
        """
        Set internal state

        :param name: Name for internal use
        :param label: Name for user-facing use
        :param worker: Coroutine function (a function defined with an
            ``async def`` syntax) that takes the job instance as a
            positional argument.

        This job finishes when `worker` returns.

        :class:`~.asyncio.CancelledError` from `worker` is ignored. Any other
        exceptions are passed to :meth:`~.JobBase.exception`.
        """
        self._name = str(name)
        self._label = str(label)
        self._worker = worker
        self._task = None

    def execute(self):
        self._task = asyncio.ensure_future(
            self._catch_cancelled_error(
                self._worker(self),
            ),
        )
        self._task.add_done_callback(self._handle_worker_done)

    async def _catch_cancelled_error(self, worker):
        try:
            return await worker
        except asyncio.CancelledError:
            _log.debug('%s: Cancelled: %r', self.name, worker)

    def _handle_worker_done(self, task):
        try:
            result = task.result()
        except BaseException as e:
            self.exception(e)
        else:
            if result:
                self.send(result)
        finally:
            self.finish()

    async def wait(self):
        """Wait for `worker`"""
        if self._task:
            await self._task
        await super().wait()

    def finish(self):
        """Cancel `worker` and finish this job"""
        if self._task:
            self._task.cancel()
        super().finish()



"""
Change or show configuration file options
"""

from .... import defaults, jobs, utils
from . import argtypes
from .base import CommandBase


def _make_pretty_option_list():
    lines = []
    for option in defaults.option_paths():
        option_type = defaults.option_type(option)
        if option_type is not str:
            lines.append(f'{option} ({option_type.__name__.lower()})')
        else:
            lines.append(option)
    return '\n  '.join(lines)


class set(CommandBase):
    """
    Change or show configuration file options

    Without any arguments, all options are listed with their current values.

    The first segment in OPTION is the file name without the extension. The
    second segment is the section name in that file. The third segment is the
    option name.

    List values are given as one argument per list item. If non-list values are
    given as multiple arguments, they are concatenated with single spaces. In
    the INI file, list items are delimited by one line break and one or more
    spaces (e.g. "\\n    ").

    Bytes values can handle units like "kB", "MB", "GiB", etc.
    """

    names = ('set',)

    description = 'options:\n  ' + _make_pretty_option_list()

    argument_definitions = {
        'OPTION': {
            'type': argtypes.option,
            'nargs': '?',
            'help': 'Option to change or show',
        },
        'VALUE': {
            'nargs': '*',
            'default': '',  # FIXME: https://bugs.python.org/issue41854
            'help': 'New value for OPTION',
            'group': 'value',
        },
        ('--reset', '-r'): {
            'action': 'store_true',
            'help': 'Reset OPTION to default value',
            'group': 'value',
        },
    }

    @utils.cached_property
    def jobs(self):
        return (
            jobs.config.SetJob(
                config=self.config,
                option=self.args.OPTION,
                # VALUE is a list. The Config class should convert lists to
                # strings (and vice versa) depending on the default type.
                value=self.args.VALUE,
                reset=self.args.reset,
            ),
        )



"""
Create screenshots from video file(s)
"""

import os
import queue

from .. import errors
from ..utils import LazyModule, daemon, fs, image, timestamp, video
from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)

DEFAULT_NUMBER_OF_SCREENSHOTS = 2

natsort = LazyModule(module='natsort', namespace=globals())


class ScreenshotsJob(JobBase):
    """
    Create screenshots from video file(s)

    This job adds the following signals to the :attr:`~.JobBase.signal`
    attribute:

        ``video_file``
            Emitted after the video file to take screenshots from is determined.
            Registered callbacks get the path to the video file as a positional
            argument.

        ``timestamps``
            Emitted after the screenshot timestamps are determined. Registered
            callbacks get the list of timestamps in the form of "H+:MM:SS" as a
            positional argument.
    """

    name = 'screenshots'
    label = 'Screenshots'

    @property
    def cache_id(self):
        """Final segment of `content_path`"""
        cache_id = [fs.basename(self._content_path)]
        if self.kwargs.get('timestamps'):
            cache_id.append(self.kwargs['timestamps'])
        if self.kwargs.get('count'):
            cache_id.append(self.kwargs['count'])
        return tuple(cache_id)

    def initialize(self, *, content_path, timestamps=(), count=0):
        """
        Set internal state

        :param str content_path: Path to file or directory (filtered through
            :func:`~.video.first_video`)
        :param timestamps: Screenshot positions in the video
        :type timestamps: sequence of "[[H+:]M+:]S+" strings or seconds
        :param count: How many screenshots to make

        If `timestamps` and `count` are not given, screenshot positions are
        picked at even intervals. If `count` is larger than the length of
        `timestamps`, more timestamps are added.
        """
        self._content_path = content_path
        self._screenshots_created = 0
        self._screenshots_total = -1
        self._video_file = ''
        self._timestamps = ()
        self.signal.add('video_file')
        self.signal.add('timestamps', record=True)
        self._screenshots_process = daemon.DaemonProcess(
            name=self.name,
            target=_screenshots_process,
            kwargs={
                'content_path' : content_path,
                'timestamps'   : timestamps,
                'count'        : count,
                'output_dir'   : self.home_directory,
                'overwrite'    : self.ignore_cache,
            },
            info_callback=self._handle_info,
            error_callback=self._handle_error,
            finished_callback=self.finish,
        )

    def execute(self):
        """
        Execute screenshot creation subprocess

        The subprocess also finds the first video and automatically picks and/or
        validates timestamps.
        """
        self._screenshots_process.start()

    def finish(self):
        """Terminate screenshot creation subprocess and finish"""
        self._screenshots_process.stop()
        super().finish()

    def _handle_info(self, info):
        if not self.is_finished:
            if info[0] == 'video_file':
                self._video_file = info[1]
                self.signal.emit('video_file', self._video_file)
            elif info[0] == 'timestamps':
                self._timestamps = tuple(info[1])
                self._screenshots_total = len(self._timestamps)
                self.signal.emit('timestamps', self._timestamps)
            elif info[0] == 'screenshot':
                self._screenshots_created += 1
                self.send(info[1])

    def _handle_error(self, error):
        if isinstance(error, BaseException):
            self.exception(error)
        else:
            self.error(error, finish=True)

    @property
    def exit_code(self):
        """`0` if all screenshots were made, `1` otherwise, `None` if unfinished"""
        if self.is_finished:
            if self.screenshots_total < 0:
                # Job is finished but _screenshots_process() never sent us
                # timestamps. That means we're either using previously cached
                # output or the job was cancelled while _screenshots_process()
                # was still initializing.
                if self.output:
                    # If we have cached output, assume the cached number of
                    # screenshots is what the user wanted because the output of
                    # unsuccessful jobs is not cached (see
                    # JobBase._write_output_cache()).
                    return 0
                else:
                    return 1
            elif len(self.output) == self.screenshots_total:
                return 0
            else:
                return 1

    @property
    def video_file(self):
        """
        Path to video file screenshots are taken from

        .. note:: This is an empty string until the subprocess is executed and
                  the ``video_file`` :attr:`~.JobBase.signal` is emitted.
        """
        return self._video_file

    @property
    def timestamps(self):
        """
        List of normalized and validated human-readable timestamps

        .. note:: This is an empty sequence until the subprocess is executed and
                  the ``timestamps`` :attr:`~.JobBase.signal` is emitted.
        """
        return self._timestamps

    @property
    def screenshots_total(self):
        """
        Total number of screenshots to make

        .. note:: This is an empty sequence until the subprocess is executed and
                  the ``timestamps`` :attr:`~.JobBase.signal` is emitted.
        """
        return self._screenshots_total

    @property
    def screenshots_created(self):
        """Total number of screenshots made so far"""
        return self._screenshots_created


def _normalize_timestamps(video_file, timestamps, count):
    """
    Return list of validated human-readable timestamps

    :params video_file: Path to video file
    :parms timestamps: Sequence of arguments for :func:`~utils.timestamp.parse`
    :parms int count: Desired number of timestamps

    :raise ValueError: if an item in `timestamps` is invalid
    :raise ContentError: if `video_file` is not a video file
    """
    total_secs = video.duration(video_file)

    timestamps_pretty = []
    for ts in timestamps:
        ts = max(0, min(total_secs, timestamp.parse(ts)))
        timestamps_pretty.append(timestamp.pretty(ts))

    if not timestamps and not count:
        count = DEFAULT_NUMBER_OF_SCREENSHOTS

    # Add more timestamps if the user didn't specify enough
    if count > 0 and len(timestamps_pretty) < count:
        # Convert timestamp strings to seconds
        timestamps = sorted(timestamp.parse(ts) for ts in timestamps_pretty)

        # Fractions of video duration
        positions = [ts / total_secs for ts in timestamps]

        # Include start and end of video
        if 0.0 not in positions:
            positions.insert(0, 0.0)
        if 1.0 not in positions:
            positions.append(1.0)

        # Insert timestamps between the two positions with the largest distance
        while len(timestamps_pretty) < count:
            pairs = [(a, b) for a, b in zip(positions, positions[1:])]
            max_distance, pos1, pos2 = max((b - a, a, b) for a, b in pairs)
            position = ((pos2 - pos1) / 2) + pos1
            timestamps_pretty.append(timestamp.pretty(total_secs * position))
            positions.append(position)
            positions.sort()

    return natsort.natsorted(timestamps_pretty)


def _screenshots_process(output_queue, input_queue,
                         content_path, timestamps, count, output_dir, overwrite):
    # Find appropriate video file if `content_path` is a directory
    try:
        video_file = video.first_video(content_path)
    except errors.ContentError as e:
        output_queue.put((daemon.MsgType.error, str(e)))
    else:
        if _shall_terminate(input_queue):
            return

        # Report video_file
        output_queue.put((daemon.MsgType.info, ('video_file', video_file)))

        # Get list of valid timestamps based on fixed timestamps and desired
        # amount of screenshots
        try:
            timestamps = _normalize_timestamps(
                video_file=video_file,
                timestamps=timestamps,
                count=count,
            )
        except (ValueError, errors.ContentError) as e:
            output_queue.put((daemon.MsgType.error, str(e)))
        else:
            # Report valid timestamps
            output_queue.put((daemon.MsgType.info, ('timestamps', timestamps)))

            for ts in timestamps:
                if _shall_terminate(input_queue):
                    return

                screenshot_file = os.path.join(
                    output_dir,
                    fs.basename(video_file) + f'.{ts}.png',
                )
                try:
                    image.screenshot(
                        video_file=video_file,
                        screenshot_file=screenshot_file,
                        timestamp=ts,
                        overwrite=overwrite,
                    )
                except errors.ScreenshotError as e:
                    output_queue.put((daemon.MsgType.error, str(e)))
                else:
                    output_queue.put((daemon.MsgType.info, ('screenshot', screenshot_file)))


def _shall_terminate(input_queue):
    try:
        typ, msg = input_queue.get_nowait()
    except queue.Empty:
        pass
    else:
        if typ == daemon.MsgType.terminate:
            return True
    return False



"""
Print ``mediainfo`` output
"""

from .... import jobs, utils
from . import argtypes
from .base import CommandBase


class mediainfo(CommandBase):
    """
    Print ``mediainfo`` output

    Directories are recursively searched for the first video file in natural
    order, e.g. "File1.mp4" comes before "File10.mp4".

    Any irrelevant leading parts in the file path are removed from the output.
    """

    names = ('mediainfo', 'mi')

    argument_definitions = {
        'CONTENT': {
            'type': argtypes.content,
            'help': 'Path to release content',
        },
    }

    @utils.cached_property
    def jobs(self):
        return (
            jobs.mediainfo.MediainfoJob(
                home_directory=self.home_directory,
                cache_directory=self.cache_directory,
                ignore_cache=self.args.ignore_cache,
                content_path=self.args.CONTENT,
            ),
        )



"""
Dump frames from video file
"""

import os
import re

from .. import errors, utils

import logging  # isort:skip
_log = logging.getLogger(__name__)

_timestamp_format = re.compile(r'^(?:(?:\d+:|)\d{2}:|)\d{2}$')


def _ffmpeg_executable():
    if utils.os_family() == 'windows':
        return 'ffmpeg.exe'
    else:
        return 'ffmpeg'


def _make_screenshot_cmd(video_file, timestamp, screenshot_file):
    return (
        _ffmpeg_executable(),
        '-y',
        '-loglevel', 'level+error',
        '-ss', str(timestamp),
        '-i', utils.video.make_ffmpeg_input(video_file),
        '-vframes', '1',
        # Use correct aspect ratio
        # https://ffmpeg.org/ffmpeg-filters.html#toc-Examples-99
        '-vf', 'scale=trunc(ih*dar):ih,setsar=1/1',
        f'file:{screenshot_file}',
    )

def screenshot(video_file, timestamp, screenshot_file, overwrite=False):
    """
    Create single screenshot from video file

    :param str video_file: Path to video file
    :param timestamp: Time location in the video
    :type timestamp: int or float or "[[H+:]MM:]SS"
    :param str screenshot_file: Path to screenshot file
    :param bool overwrite: Whether to overwrite `screenshot_file` if it exists

    :raise ScreenshotError: if something goes wrong
    """
    try:
        utils.fs.assert_file_readable(video_file)
    except errors.ContentError as e:
        raise errors.ScreenshotError(e)

    if isinstance(timestamp, str):
        if not _timestamp_format.match(timestamp):
            raise errors.ScreenshotError(f'Invalid timestamp: {timestamp!r}')
    elif not isinstance(timestamp, (int, float)):
        raise errors.ScreenshotError(f'Invalid timestamp: {timestamp!r}')

    if not overwrite and os.path.exists(screenshot_file):
        _log.debug('Screenshot already exists: %s', screenshot_file)
        return screenshot_file

    duration = utils.video.duration(video_file)
    if duration <= utils.timestamp.parse(timestamp):
        raise errors.ScreenshotError(
            f'Timestamp is after video end ({utils.timestamp.pretty(duration)}): '
            + utils.timestamp.pretty(timestamp)
        )

    cmd = _make_screenshot_cmd(video_file, timestamp, screenshot_file)
    output = utils.subproc.run(cmd, ignore_errors=True, join_stderr=True)
    if not os.path.exists(screenshot_file):
        raise errors.ScreenshotError(
            f'{video_file}: Failed to create screenshot at {timestamp}: {output}'
        )
    else:
        return screenshot_file


def _make_resize_cmd(image_file, dimensions, resized_file):
    return (
        _ffmpeg_executable(),
        '-y',
        '-loglevel', 'level+error',
        '-i', f'file:{image_file}',
        '-vf', f'scale={dimensions}',
        f'file:{resized_file}',
    )

def resize(image_file, width=None, height=None):
    try:
        utils.fs.assert_file_readable(image_file)
    except errors.ContentError as e:
        raise errors.ImageResizeError(e)

    if width and width < 1:
        raise errors.ImageResizeError(f'Width must be greater than zero: {width}')
    elif height and height < 1:
        raise errors.ImageResizeError(f'Height must be greater than zero: {height}')
    elif width and height:
        dimensions = f'w={int(width)}:h={int(height)}'
        extension = f'.{width}x{height}.'
    elif width:
        dimensions = f'w={int(width)}:h=-1'
        extension = f'.{width}x-1.'
    elif height:
        dimensions = f'w=-1:h={int(height)}'
        extension = f'.-1x{height}.'
    else:
        return image_file

    target_path = (
        utils.fs.strip_extension(image_file)
        + extension
        + utils.fs.file_extension(image_file)
    )
    _log.debug('Resizing to %r: %r', dimensions, image_file)
    _log.debug('Resize target: %r', target_path)

    cmd = _make_resize_cmd(image_file, dimensions, target_path)
    output = utils.subproc.run(cmd, ignore_errors=True, join_stderr=True)
    if not os.path.exists(target_path):
        raise errors.ImageResizeError(
            f'{image_file}: Failed to resize: {output}'
        )
    else:
        return target_path



"""
Upload images to image hosting service
"""

from .... import jobs, utils
from . import argtypes
from .base import CommandBase


class upload_images(CommandBase):
    """Upload images to image hosting service"""

    names = ('upload-images', 'ui')

    argument_definitions = {
        'IMAGEHOST': {
            'type': argtypes.imghost,
            'help': ('Case-insensitive name of image hosting service\n'
                     'Supported services: ' + ', '.join(utils.imghosts.imghost_names())),
        },
        'IMAGE': {
            'nargs': '+',
            'help': 'Path to image file',
        },
    }

    @utils.cached_property
    def jobs(self):
        return (
            jobs.imghost.ImageHostJob(
                home_directory=self.home_directory,
                cache_directory=self.cache_directory,
                ignore_cache=self.args.ignore_cache,
                imghost=utils.imghosts.imghost(
                    name=self.args.IMAGEHOST,
                    **self.config['imghosts'][self.args.IMAGEHOST],
                ),
                enqueue=self.args.IMAGE,
            ),
        )



"""
API for BitTorrent clients
"""

from .. import CaseInsensitiveString, subclasses, submodules
from .base import ClientApiBase


def clients():
    """Return list of :class:`.ClientApiBase` subclasses"""
    return subclasses(ClientApiBase, submodules(__package__))


def client(name, **kwargs):
    """
    Create :class:`.ClientApiBase` instance

    :param str name: Name of the client. A subclass of :class:`.ClientApiBase`
        with the same :attr:`~.ClientApiBase.name` must exist in one of this
        package's submodules.
    :param kwargs: All keyword arguments are passed to the subclass specified by
        `name`

    :raise ValueError: if no matching subclass can be found

    :return: :class:`.ClientApiBase` instance
    """
    for client in clients():
        if client.name == name:
            return client(**kwargs)
    raise ValueError(f'Unsupported client: {name}')


def client_names():
    """Return sequence of valid `name` arguments for :func:`.client`"""
    return sorted(CaseInsensitiveString(cls.name) for cls in clients())



"""
Connect the user interface to the engine room
"""

from .base import JobBase, QueueJobBase  # isort:skip
from . import (config, custom, dialog, imghost, mediainfo, release_name, scene,
               screenshots, submit, torrent, webdb)



"""
Open a URL in a web browser
"""

import asyncio
import functools

from . import LazyModule

import logging  # isort:skip
_log = logging.getLogger(__name__)

webbrowser = LazyModule(module='webbrowser', namespace=globals())


def open(url):
    """Attempt to open URL in default web browser"""
    # Don't block while the browser is being started
    wrapper = functools.partial(webbrowser.open_new_tab, str(url))
    loop = asyncio.get_event_loop()
    loop.run_in_executor(None, wrapper)



"""
Query online databases like IMDb
"""

import asyncio
import collections
from time import monotonic as time_monotonic

from .. import errors
from ..utils import fs, webdbs
from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class SearchWebDbJob(JobBase):
    """
    Ask user to select a specific search result from an internet database

    This job adds the following signals to the :attr:`~.JobBase.signal`
    attribute:

        ``search_results``
            Emitted after new search results are available. Registered callbacks
            get a sequence of :class:`~.utils.webdbs.common.SearchResult`
            instances as a positional argument.

        ``searching_status``
            Emitted when a new search is started and when new results are
            available. Registered callbacks get `True` for "search started" or
            `False` for "search ended" as a positional argument.

        ``info_updated``
            Emitted when additional information is available. Registered
            callbacks are called for each piece of information and get ``key``
            and ``value`` as positional arguments. ``key`` is an attribute of a
            :class:`~.utils.webdbs.common.SearchResult` object that returns a
            coroutine function. ``value`` is the return value of that coroutine
            function.
    """

    @property
    def name(self):
        return f'{self._db.name}-id'

    @property
    def label(self):
        return f'{self._db.label} ID'

    @property
    def cache_id(self):
        """Final segment of `content_path` and database :attr:`~.WebDbApiBase.name`"""
        return fs.basename(self._content_path)

    @property
    def query(self):
        """:class:`~.webdbs.common.Query` instance"""
        return self._query

    def initialize(self, *, db, content_path):
        """
        Set internal state

        :param WebDbApiBase db: Return value of :func:`.utils.webdbs.webdb`
        :param str content_path: Path or name of the release (doesn't have to
            exist)
        """
        assert isinstance(db, webdbs.WebDbApiBase), f'Not a WebDbApiBase: {db!r}'
        self._db = db
        self._content_path = content_path
        self._query = webdbs.Query.from_path(content_path)
        self._is_searching = False

        self.signal.add('search_results')
        self.signal.add('searching_status')
        self.signal.add('info_updated')

        self._searcher = _Searcher(
            search_coro=self._db.search,
            results_callback=self._handle_search_results,
            searching_callback=self._handle_searching_status,
            error_callback=self.warn,
            exception_callback=self.exception,
        )
        self._info_updater = _InfoUpdater(
            error_callback=self.warn,
            exception_callback=self.exception,
            targets={
                'id': self._make_update_info_func('id'),
                'summary': self._make_update_info_func('summary'),
                'title_original': self._make_update_info_func('title_original'),
                'title_english': self._make_update_info_func('title_english'),
                'keywords': self._make_update_info_func('keywords'),
                'director': self._make_update_info_func('director'),
                'cast': self._make_update_info_func('cast'),
                'countries': self._make_update_info_func('countries'),
            },
        )

    def _make_update_info_func(self, key):
        return lambda value: self._update_info(key, value)

    def _update_info(self, attr, value):
        self.signal.emit('info_updated', attr, value)

    def execute(self):
        """Search for initial query"""
        # It is important NOT to do this in initialize() because the window
        # between initialize() and execute() is used to register callbacks.
        self._searcher.search(self._query)

    async def wait(self):
        """Wait for any running internal coroutines"""
        # Raise any exceptions and avoid warnings about unawaited tasks
        await asyncio.gather(
            self._searcher.wait(),
            self._info_updater.wait(),
        )
        await super().wait()

    def finish(self):
        """Cancel any running internal coroutines and finish"""
        self._searcher.cancel()
        self._info_updater.cancel()
        super().finish()

    def search(self, query):
        """Make a new search request after cancelling any currently ongoing request"""
        if not self.is_finished:
            self._query = webdbs.Query.from_string(query)
            self._searcher.search(self._query)
            self.clear_warnings()

    def _handle_searching_status(self, is_searching):
        self._is_searching = bool(is_searching)
        self.signal.emit('searching_status', self._is_searching)

    @property
    def is_searching(self):
        """Whether a search request is currently being made"""
        return self._is_searching

    def _handle_search_results(self, results):
        if results:
            self._info_updater.set_result(results[0])
        else:
            self._info_updater.set_result(None)
        self.signal.emit('search_results', results)

    def result_focused(self, result):
        """
        Must be called by the UI when the user focuses a different search result

        :param SearchResult result: Focused search result or `None` if there are
            no results
        """
        self._info_updater.set_result(result)

    def result_selected(self, result):
        """
        Must be called by the UI when the user accepts a search result

        :param SearchResult result: Focused search result or `None` if there are
            no results of if the user refused every result
        """
        if not self.is_searching:
            if result is not None:
                self.send(str(result.id))
            self.finish()


class _Searcher:
    def __init__(self, search_coro, results_callback, searching_callback,
                 error_callback, exception_callback):
        self._search_coro = search_coro
        self._results_callback = results_callback
        self._searching_callback = searching_callback
        self._error_callback = error_callback
        self._exception_callback = exception_callback
        self._previous_search_time = 0
        self._search_task = None

    async def wait(self):
        if self._search_task:
            try:
                await self._search_task
            except asyncio.CancelledError:
                pass

    def cancel(self):
        if self._search_task:
            self._search_task.cancel()

    def _handle_search_task(self, task):
        try:
            if task.exception():
                self._exception_callback(task.exception())
        except asyncio.CancelledError:
            pass

    def search(self, query):
        """
        Schedule new query

        :param query: Query to make
        :type query: :class:`~.utils.webdbs.Query`
        """
        if self._search_task:
            self._search_task.cancel()
            self._search_task = None
        self._search_task = asyncio.ensure_future(self._search(query))
        self._search_task.add_done_callback(self._handle_search_task)

    async def _search(self, query):
        self._results_callback(())
        self._searching_callback(True)
        await self._delay()
        results = ()
        try:
            results = await self._search_coro(query)
        except errors.RequestError as e:
            self._error_callback(e)
        finally:
            self._searching_callback(False)
            self._results_callback(results)

    _min_seconds_between_searches = 1

    async def _delay(self):
        cur_time = time_monotonic()
        time_since_prev_search = cur_time - self._previous_search_time
        if time_since_prev_search <= self._min_seconds_between_searches:
            remaining_delay = self._min_seconds_between_searches - time_since_prev_search
            await asyncio.sleep(remaining_delay)
        self._previous_search_time = time_monotonic()


class _InfoUpdater:
    def __init__(self, targets, error_callback, exception_callback):
        super().__init__()
        # `targets` maps names of SearchResult attributes to callbacks that get
        # the value of each attribute. For example, {"title": handle_title}
        # means: For each search_result, call handle_title(search_result.title).
        # Values of SearchResult attributes may also be coroutine functions that
        # return the actual value. In that case, call
        # handle_title(await search_result.title()).
        self._targets = targets
        self._error_callback = error_callback
        self._exception_callback = exception_callback
        # SearchResult instance or None
        self._result = None
        self._update_task = None

    async def wait(self):
        if self._update_task:
            try:
                await self._update_task
            except asyncio.CancelledError:
                pass

    def cancel(self):
        if self._update_task:
            self._update_task.cancel()

    def _handle_update_task(self, task):
        try:
            if task.exception():
                self._exception_callback(task.exception())
        except asyncio.CancelledError:
            pass

    def set_result(self, result):
        self.cancel()
        self._result = result

        if not self._result:
            for callback in self._targets.values():
                callback('')
        else:
            # Schedule calls of SearchResult attributes that are coroutine
            # functions
            self._update_task = asyncio.ensure_future(self._update())
            self._update_task.add_done_callback(self._handle_update_task)

            # Update plain, non-callable values immediately
            if result is not None:
                for attr, callback in self._targets.items():
                    value = getattr(result, attr)
                    if not callable(value):
                        callback(self._value_as_string(value))

    async def _update(self):
        tasks = []
        for attr, callback in self._targets.items():
            value = getattr(self._result, attr)
            if callable(value):
                tasks.append(self._call_callback(
                    callback=callback,
                    value_getter=value,
                    cache_key=(self._result.id, attr),
                ))
        await asyncio.gather(*tasks)

    _cache = {}
    _delay_between_updates = 0.5

    async def _call_callback(self, callback, value_getter, cache_key):
        cached_value = self._cache.get(cache_key, None)
        if cached_value is not None:
            callback(cached_value)
        else:
            callback('Loading...')
            await asyncio.sleep(self._delay_between_updates)
            try:
                value = await value_getter()
            except errors.RequestError as e:
                callback('')
                self._error_callback(e)
            else:
                value_str = self._value_as_string(value)
                self._cache[cache_key] = value_str
                callback(value_str)

    @staticmethod
    def _value_as_string(value):
        if not isinstance(value, str) and isinstance(value, collections.abc.Iterable):
            return ', '.join(str(v) for v in value)
        else:
            return str(value)



"""
CLI argument types

All types return normalized values and raise ValueError for invalid values.
"""

import enum
import re


class Bytes(int):
    """:class:`int` subclass that interprets units and unit prefixes"""

    _regex = re.compile(r'^(\d+(?:\.\d+|)) ?([a-zA-Z]{,3})$')
    _multipliers = {
        '': 1,
        'k': 1000,
        'M': 1000**2,
        'G': 1000**3,
        'T': 1000**4,
        'P': 1000**5,
        'Ki': 1024,
        'Mi': 1024**2,
        'Gi': 1024**3,
        'Ti': 1024**4,
        'Pi': 1024**5,
    }

    @classmethod
    def from_string(cls, string):
        match = cls._regex.search(string)
        if not match:
            raise ValueError(f'Invalid size: {string}')
        else:
            number = match.group(1)
            unit = match.group(2)
            if unit and unit[-1] == 'B':
                unit = unit[:-1]
            try:
                multiplier = cls._multipliers[unit]
            except KeyError:
                raise ValueError(f'Invalid unit: {unit}')
            else:
                return cls(int(float(number) * multiplier))

    def __new__(cls, value):
        if isinstance(value, str):
            return cls.from_string(value)
        else:
            return super().__new__(cls, value)

    def __str__(self):
        decimal_multipliers = (
            (prefix, multiplier)
            for prefix, multiplier in reversed(tuple(self._multipliers.items()))
            if len(prefix) == 1
        )
        binary_multipliers = (
            (prefix, multiplier)
            for prefix, multiplier in reversed(tuple(self._multipliers.items()))
            if len(prefix) == 2
        )

        def get_string(multipliers):
            for prefix, multiplier in multipliers:
                if self >= multiplier:
                    return f'{self / multiplier:.2f}'.rstrip('0').rstrip('.') + f' {prefix}B'
            return f'{int(self)} B'

        def number_of_decimal_places(number):
            string = str(''.join(c for c in str(number) if c in '1234567890.'))
            if '.' in string:
                return len(string.split('.', maxsplit=1)[1])
            else:
                return 0

        decimal_string = get_string(decimal_multipliers)
        binary_string = get_string(binary_multipliers)
        sorted_strings = sorted((decimal_string, binary_string),
                                key=number_of_decimal_places)
        return sorted_strings[0]


class ReleaseType(enum.Enum):
    """
    Enum with the values ``movie``, ``season``, ``episode`` and
    ``unknown``

    ``series`` is an alias for ``season``.

    All values are truthy except for ``unknown``.
    """

    movie = 'movie'
    season = 'season'
    series = 'season'
    episode = 'episode'
    unknown = 'unknown'

    def __bool__(self):
        return self is not self.unknown

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return f'{type(self).__name__}.{self.value}'


class SceneCheckResult(enum.Enum):
    """
    Enum with the values ``true``, ``false``, ``renamed``, ``altered`` and
    ``unknown``

    All values are falsy except for ``true``.
    """

    true = 'true'
    false = 'false'
    renamed = 'renamed'
    altered = 'altered'
    unknown = 'unknown'

    def __bool__(self):
        return self is self.true

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return f'{type(self).__name__}.{self.value}'



"""
Base class for client APIs
"""

import abc

from ... import errors


class ClientApiBase(abc.ABC):
    """
    Base class for all BitTorrent client APIs
    """

    @property
    @abc.abstractmethod
    def name(self):
        """Name of the BitTorrent client"""

    @abc.abstractmethod
    def add_torrent(self, torrent_path, download_path=None):
        """
        Add torrent to client

        :param str torrent_path: Path to torrent file
        :param str download_path: Path to the directory where the torrent's
            content is or `None` to use the client's default

        :raise TorrentError: if adding `torrent_path` fails

        :return: Info hash of the added torrent
        """

    @staticmethod
    def read_torrent_file(torrent_path):
        """
        Return bytes from torrent file

        :param str torrent_path: Path to torrent file

        :raise TorrentError: if reading from `torrent_path` fails
        """
        try:
            with open(torrent_path, 'rb') as f:
                return f.read()
        except OSError as e:
            raise errors.TorrentError(f'{torrent_path}: {e.strerror}')



"""
Check if scene release was altered
"""

import asyncio
import os
import re

from ... import errors, utils
from ..types import ReleaseType, SceneCheckResult
from . import common, predb, srrdb
from .find import SceneQuery

import logging  # isort:skip
_log = logging.getLogger(__name__)

_predb = predb.PreDbApi()
_srrdb = srrdb.SrrDbApi()

_abbreviated_scene_filename_regexs = (
    # Match names with group in front
    re.compile(r'^[a-z0-9]+-[a-z0-9_\.-]+?(?!-[a-z]{2,})\.(?:mkv|avi)$'),
    # Match "ttl.720p-group.mkv"
    re.compile(r'^[a-z0-9]+[\.-]\d{3,4}p-[a-z]{2,}\.(?:mkv|avi)$'),
    # Match "GR0UP1080pTTL.mkv"
    re.compile(r'^[a-zA-Z0-9]+\d{3,4}p[a-zA-Z]+\.(?:mkv|avi)$'),
)


def assert_not_abbreviated_filename(filepath):
    """
    Raise :class:`~.errors.SceneError` if `filepath` points to an abbreviated
    scene release file name like ``abd-mother.mkv``
    """
    filename = os.path.basename(filepath)
    for regex in _abbreviated_scene_filename_regexs:
        if regex.search(filename):
            raise errors.SceneError(
                f'Provide parent directory of abbreviated scene file: {filename}'
            )


async def is_scene_release(release):
    """
    Check if `release` is a scene release or not

    :param release: Release name, path to release or
        :class:`~.release.ReleaseInfo` instance

    :return: :class:`~.types.SceneCheckResult`
    """
    if isinstance(release, str):
        release_info = utils.release.ReleaseInfo(release)
    else:
        release_info = release

    # NOTE: Simply searching for the group does not work because some scene
    #       groups make non-scene releases and vice versa.
    #       Examples:
    #         - Prospect.2018.720p.BluRay.DD5.1.x264-LoRD
    #         - How.The.Grinch.Stole.Christmas.2000.720p.BluRay.DTS.x264-EbP
    query = SceneQuery.from_release(release_info)
    results = await _predb.search(query)
    if results:
        # Do we have enough information to pinpoint a single release?
        needed_keys = common.get_needed_keys(release_info)
        if not needed_keys:
            # If we don't know how to identify a release uniquely, we are not in
            # a position to make any claims.
            return SceneCheckResult.unknown
        elif all(release_info[k] for k in needed_keys):
            return SceneCheckResult.true
        else:
            return SceneCheckResult.unknown

    # If this is a file like "abd-mother.mkv" without a properly named parent
    # directory and we didn't find it above, it's possibly a scene release, but
    # we can't be sure.
    try:
        assert_not_abbreviated_filename(release_info.path)
    except errors.SceneError:
        return SceneCheckResult.unknown

    return SceneCheckResult.false


async def release_files(release_name):
    """
    Map release file names to file information

    This function uses :class:`~.predb.PreDbApi` for searching and
    :class:`~.srrdb.SrrDbApi` to get the file information.

    :param str release_name: Exact name of the release
    """
    files = await _srrdb.release_files(release_name)
    if files:
        return files
    else:
        _log.debug('No such release: %r', release_name)
        files = {}

    # If scene released "Foo.S01E0{1,2,3,...}.720p.BluRay-BAR" and we're
    # searching for "Foo.S01.720p.BluRay-BAR", we might not get any results. But
    # we can get release names of individual episodes by searching for the
    # season pack, and then we can call release_files() for each episode.
    release_info = utils.release.ReleaseInfo(release_name)
    if release_info['type'] is ReleaseType.season:
        query = SceneQuery.from_release(release_info)
        results = await _predb.search(query)
        if results:
            files = await asyncio.gather(
                *(_srrdb.release_files(result) for result in results)
            )
            # Join sequence of dictionaries into single dictionary
            files = {fname:f for files_ in files for fname,f in files_.items()}
            _log.debug('Season pack from multiple episode releases: %r', files)

    # If scene released season pack (or any other multi-file thing) and we're
    # searching for a single episode, we might not get any results. Search for
    # the season pack to get all files.
    elif release_info['type'] is ReleaseType.episode:
        # Remove single episodes from seasons
        release_info['episodes'].remove_specific_episodes()
        results = await _predb.search(SceneQuery.from_release(release_info))
        if len(results) == 1:
            _log.debug('Getting files from single result: %r', results[0])
            files = await _srrdb.release_files(results[0])

    # Go through all files and find the exact release name we're looking for.
    # Don't do this exclusively for episodes because not all multi-file releases
    # are a list of episodes (e.g. extras may not contain any "Exx").
    for file_name, file_info in files.items():
        if utils.fs.strip_extension(release_name) == utils.fs.strip_extension(file_name):
            files = {file_name: file_info}
            _log.debug('Single file from season pack release: %r', files)
            break

    return files


async def verify_release_name(content_path, release_name):
    """
    Check if release was renamed

    :param content_path: Path to release file or directory
    :param release_name: Known exact release name, e.g. from :func:`search`
        results

    `content_path` is ok if its last segment is equal to `release_name` (file
    extension is ignored) or equal to one of the files from the release
    specified by `release_name`.

    :raise SceneRenamedError: if release was renamed
    :raise SceneError: if release name is not a scene release or if
        `content_path` points to an abbreviated file name
        (e.g. "abd-mother.mkv")
    """
    assert_not_abbreviated_filename(content_path)
    if not await is_scene_release(release_name):
        raise errors.SceneError(f'Not a scene release: {release_name}')

    content_filename = utils.fs.basename(content_path)
    content_release_name = utils.fs.strip_extension(content_filename)

    if content_release_name == release_name:
        return

    files = await release_files(release_name)
    if content_filename in files:
        return

    raise errors.SceneRenamedError(
        original_name=release_name,
        existing_name=content_release_name,
    )


async def verify_release_files(content_path, release_name):
    """
    Check if files released by scene have the correct size

    :param content_path: Path to release file or directory
    :param release_name: Known exact release name, e.g. from :func:`search`
        results

    The return value is a sequence of :class:`~.errors.SceneError` exceptions.
    For every file that is part of the scene release specified by
    `release_name`, include:

        * :class:`~.errors.SceneFileSizeError` if it has the wrong size
        * :class:`~.errors.SceneMissingInfoError` if file information is missing
        * :class:`~.errors.SceneError` if release name is not a scene release or
          if `content_path` points to an abbreviated file name
          (e.g. "abd-mother.mkv")
    """
    exceptions = []
    try:
        assert_not_abbreviated_filename(content_path)
    except errors.SceneError as e:
        exceptions.append(e)
    if not await is_scene_release(release_name):
        exceptions.append(errors.SceneError(f'Not a scene release: {release_name}'))
    if exceptions:
        return tuple(exceptions)

    fileinfos = await release_files(release_name)

    def get_release_filesize(filename):
        return fileinfos.get(filename, {}).get('size', None)

    # Map file paths to expected file sizes
    if os.path.isdir(content_path):
        exp_filesizes = {filepath: get_release_filesize(utils.fs.basename(filepath))
                         for filepath in utils.fs.file_list(content_path)}
    else:
        if len(fileinfos) == 1:
            filename = tuple(fileinfos)[0]
            exp_filesize = get_release_filesize(filename)
            exp_filesizes = {
                # Title.2015.720p.BluRay.x264-FOO.mkv
                content_path: exp_filesize,
                # Title.2015.720p.BluRay.x264-FOO/foo-title.mkv
                os.path.join(utils.fs.strip_extension(content_path), filename): exp_filesize,
            }
        else:
            filename = utils.fs.basename(content_path)
            exp_filesizes = {content_path: get_release_filesize(filename)}

    # Compare expected file sizes to actual file sizes
    _log.debug('file sizes: %r', exp_filesizes)
    for filepath, exp_size in exp_filesizes.items():
        filename = utils.fs.basename(filepath)
        actual_size = utils.fs.file_size(filepath)
        _log.debug('Checking file size: %s: %r ?= %r', filename, actual_size, exp_size)
        if exp_size is None:
            _log.debug('No info: %s', filename)
            exceptions.append(errors.SceneMissingInfoError(filename))
        elif actual_size is not None:
            if actual_size != exp_size:
                _log.debug('Wrong size: %s', filename)
                exceptions.append(
                    errors.SceneFileSizeError(
                        filename=filename,
                        original_size=exp_size,
                        existing_size=actual_size,
                    )
                )
            else:
                _log.debug('Correct size: %s', filename)
        else:
            _log.debug('No such file: %s', filename)

    return tuple(e for e in exceptions if e)


async def verify_release(content_path, release_name):
    """
    Combine :func:`is_scene_release`, :func:`verify_release_name` and
    :func:`verify_release_files`

    It is safe to pass non-scene releases and will result in the return value
    `(SceneCheckResult.false, ())`.

    :param content_path: Path to release file or directory
    :param release_name: Known exact release name, e.g. from :func:`search`
        results

    :return: :class:`~.types.SceneCheckResult` enum from
        :func:`is_scene_release` and sequence of :class:`~.errors.SceneError`
        exceptions from :func:`verify_release_name` and
        :func:`verify_release_files`
    """
    # Don't allow abbreviated scene release files, e.g. "abd-mother.mkv"
    try:
        assert_not_abbreviated_filename(content_path)
    except errors.SceneError as e:
        return SceneCheckResult.unknown, (e,)

    # Stop other checks if this is not a scene release
    is_scene = await is_scene_release(release_name)
    if not is_scene:
        return is_scene, ()

    # Combine exceptions from verify_release_name() and verify_release_files()
    exceptions = []
    try:
        await verify_release_name(content_path, release_name)
    except errors.SceneError as e:
        exceptions.append(e)
    finally:
        exceptions.extend(await verify_release_files(content_path, release_name))
        return is_scene, tuple(exceptions)



"""
Abstract base class for online databases
"""

import abc
import asyncio


class WebDbApiBase(abc.ABC):
    """
    Base class for all web DB APIs

    Because not all DBs provide all information, methods that take an `id`
    argument may raise :class:`NotImplementedError`.
    """

    @property
    @abc.abstractmethod
    def name(self):
        """Unique name of this DB"""

    @property
    @abc.abstractmethod
    def label(self):
        """User-facing name of this DB"""

    @abc.abstractmethod
    async def search(self, query):
        """
        Search DB

        :param query: :class:`~.common.Query` instance

        :return: List of :class:`~.common.SearchResult` instances
        """

    @abc.abstractmethod
    async def cast(self, id):
        """Return list of cast names"""

    @abc.abstractmethod
    async def countries(self, id):
        """Return list of country names"""

    @abc.abstractmethod
    async def creators(self, id):
        """Return list of creator names (usually empty for movies and episodes)"""

    @abc.abstractmethod
    async def directors(self, id):
        """Return list of director names (usually empty for series)"""

    @abc.abstractmethod
    async def keywords(self, id):
        """Return list of keywords, e.g. genres"""

    @abc.abstractmethod
    async def poster_url(self, id):
        """Return URL of poster image or `None`"""

    @abc.abstractmethod
    async def rating(self, id):
        """Return rating as a number or `None`"""

    @property
    @abc.abstractmethod
    async def rating_min(self):
        """Minimum :meth:`rating` value"""

    @property
    @abc.abstractmethod
    async def rating_max(self):
        """Maximum :meth:`rating` value"""

    @abc.abstractmethod
    async def summary(self, id):
        """Return short plot description"""

    @abc.abstractmethod
    async def title_english(self, id):
        """Return English title if different from original title or empty string"""

    @abc.abstractmethod
    async def title_original(self, id):
        """Return original title (e.g. non-English) or empty string"""

    @abc.abstractmethod
    async def type(self, id):
        """Return :class:`~.types.ReleaseType`"""

    @abc.abstractmethod
    async def url(self, id):
        """Return URL for `id`"""

    @abc.abstractmethod
    async def year(self, id):
        """Return release year or empty string"""

    async def gather(self, id, *methods):
        """
        Fetch information concurrently

        :param id: Valid ID for this DB
        :param methods: Names of coroutine methods of this class
        :type methods: sequence of :class:`str`

        :return: Dictionary that maps `methods` to return values
        """
        corofuncs = (getattr(self, method) for method in methods)
        awaitables = (corofunc(id) for corofunc in corofuncs)
        results = await asyncio.gather(*awaitables)
        dct = {'id': id}
        # "The order of result values corresponds to the order of awaitables in `aws`."
        # https://docs.python.org/3/library/asyncio-task.html#running-tasks-concurrently
        dct.update((method, result) for method, result in zip(methods, results))
        return dct



"""
Search online database like IMDb to get an ID
"""

from .... import jobs, utils
from . import argtypes
from .base import CommandBase


class webdb_search(CommandBase):
    """
    Search online database like IMDb to get an ID

    Pressing ``Enter`` searches for the current query. Pressing ``Enter`` again without
    changing the query selects the focused search result.

    The focused search result can be opened in the default web browser with
    ``Alt-Enter``.
    """

    names = ('id',)

    argument_definitions = {
        'DB': {
            'type': argtypes.webdb,
            'help': ('Case-insensitive database name\n'
                     'Supported databases: ' + ', '.join(utils.webdbs.webdb_names())),
        },
        'RELEASE': {
            'type': argtypes.release,
            'help': 'Release name or path to release content',
        },
    }

    @utils.cached_property
    def jobs(self):
        return (
            jobs.webdb.SearchWebDbJob(
                home_directory=self.home_directory,
                cache_directory=self.cache_directory,
                ignore_cache=self.args.ignore_cache,
                content_path=self.args.RELEASE,
                db=utils.webdbs.webdb(self.args.DB),
            ),
        )



from . import constants, trackers, utils

defaults = {
    'config': {
        'main': {
            'cache_directory': constants.CACHE_DIRPATH,
            'max_cache_size': utils.types.Bytes.from_string('20 MB'),
        },
    },

    'trackers': {
        tracker.name: tracker.TrackerConfig()
        for tracker in trackers.trackers()
    },

    'imghosts': {
        'dummy': {},
        'imgbox': {},
    },

    'clients': {
        'dummy': {},
        'transmission': {
            'url': 'http://localhost:9091/transmission/rpc',
            'username': '',
            'password': '',
        },
    },
}
"""Defaults for configuration options"""


def option_paths():
    """Tuple of configuration option paths (`<section>.<subsection>.<option>`)"""

    def _get_option_paths(cfg, parents=()):
        paths = []
        for k, v in cfg.items():
            k_parents = parents + (k,)
            if isinstance(v, dict):
                paths.extend(_get_option_paths(v, parents=k_parents))
            else:
                paths.append('.'.join(k_parents))
        return tuple(sorted(paths))

    return _get_option_paths(defaults)


def option_type(option_path):
    filename, section, option = option_path.split('.', maxsplit=2)
    return type(defaults[filename][section][option])



"""
Client API for the Transmission daemon
"""

import base64
import json
import os

from ... import errors
from .. import http
from .base import ClientApiBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


DEFAULT_URL = 'http://localhost:9091/transmission/rpc'
AUTH_ERROR_CODE = 401
CSRF_ERROR_CODE = 409
CSRF_HEADER = 'X-Transmission-Session-Id'


class TransmissionClientApi(ClientApiBase):
    """
    Transmission daemon API

    Reference: https://github.com/transmission/transmission/blob/master/extras/rpc-spec.txt

    :param str username: Username used for authentication
    :param str password: Password used for authentication
    :param str url: Transmission RPC URL, defaults to
        "http://localhost:9091/transmission/rpc"
    """

    name = 'transmission'

    def __init__(self, username='', password='', url=None):
        self._username = username
        self._password = password
        self._url = url or DEFAULT_URL
        self._headers = {'Content-Type': 'application/json'}

    async def _request(self, data):
        if self._username or self._password:
            auth = (self._username, self._password)
        else:
            auth = None

        try:
            response = await http.post(
                url=self._url,
                headers=self._headers,
                auth=auth,
                data=data,
            )
        except errors.RequestError as e:
            if e.status_code == CSRF_ERROR_CODE:
                # Send same request again with CSRF header
                self._headers[CSRF_HEADER] = e.headers[CSRF_HEADER]
                return await self._request(data)
            elif e.status_code == AUTH_ERROR_CODE:
                raise errors.TorrentError('Authentication failed')
            else:
                raise errors.TorrentError(e)

        try:
            return response.json()
        except errors.RequestError as e:
            raise errors.TorrentError(e)

    async def add_torrent(self, torrent_path, download_path=None):
        torrent_data = str(
            base64.b64encode(self.read_torrent_file(torrent_path)),
            encoding='ascii',
        )
        request = {
            'method' : 'torrent-add',
            'arguments' : {
                'metainfo': torrent_data,
            },
        }
        if download_path:
            request['arguments']['download-dir'] = str(os.path.abspath(download_path))

        info = await self._request(json.dumps(request))
        arguments = info.get('arguments', {})
        if 'torrent-added' in arguments:
            return arguments['torrent-added']['hashString']
        elif 'torrent-duplicate' in arguments:
            return arguments['torrent-duplicate']['hashString']
        elif 'result' in info:
            raise errors.TorrentError(str(info['result']).capitalize())
        else:
            raise RuntimeError(f'Unexpected response: {info}')



"""
API for imdb.com
"""

import functools
import re
import string

from .. import html, http
from ..types import ReleaseType
from . import common
from .base import WebDbApiBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class ImdbApi(WebDbApiBase):
    """API for imdb.com"""

    name = 'imdb'
    label = 'IMDb'
    _url_base = 'https://imdb.com'

    _soup_cache = {}

    async def _get_soup(self, path, params={}):
        cache_id = (path, tuple(sorted(params.items())))
        if cache_id in self._soup_cache:
            return self._soup_cache[cache_id]
        text = await http.get(
            url=f'{self._url_base}/{path}',
            params=params,
            cache=True,
        )
        self._soup_cache[cache_id] = html.parse(text)
        return self._soup_cache[cache_id]

    _title_types = {
        ReleaseType.movie: 'feature,tv_movie,documentary,short,video,tv_short',
        ReleaseType.season: 'tv_series,tv_miniseries',
        # Searching for single episodes is currently not supported
        ReleaseType.episode: 'tv_series,tv_miniseries',
    }

    async def search(self, query):
        _log.debug('Searching IMDb for %s', query)
        if not query.title:
            return []

        path = 'search/title/'
        params = {'title': query.title_normalized}
        if query.type is not ReleaseType.unknown:
            params['title_type'] = self._title_types[query.type]
        if query.year is not None:
            params['release_date'] = f'{query.year}-01-01,{query.year}-12-31'

        soup = await self._get_soup(path, params=params)
        items = soup.find_all('div', class_='lister-item-content')
        results = [_ImdbSearchResult(soup=item, imdb_api=self)
                   for item in items]
        return results

    _person_url_path_regex = re.compile(r'(/name/nm\d+)')

    def _get_persons(self, tag):
        a_tags = tag.find_all('a', href=self._person_url_path_regex)
        persons = []
        for a_tag in a_tags:
            if a_tag.string:
                name = a_tag.string.strip()
                url_path = self._person_url_path_regex.match(a_tag["href"]).group(1)
                url = f'{self._url_base.rstrip("/")}/{url_path.lstrip("/")}'
                persons.append(common.Person(name, url))
        return tuple(persons)

    async def cast(self, id):
        soup = await self._get_soup(f'title/{id}')
        cast = []
        cast_tag = soup.find(class_='cast_list')
        if cast_tag:
            tr_tags = cast_tag.find_all('tr')
            for tr_tag in tr_tags:
                cast.extend(self._get_persons(tr_tag))
        return tuple(cast)

    async def countries(self, id):
        soup = await self._get_soup(f'title/{id}')
        details_tag = soup.find(id='titleDetails')
        countries_tag = details_tag.find(string='Country:').parent.parent
        countries = []
        for country_tag in countries_tag.find_all('a'):
            countries.append(''.join(country_tag.stripped_strings))
        return tuple(countries)

    async def creators(self, id):
        soup = await self._get_soup(f'title/{id}')
        creators = []
        for tag in soup.find_all(class_='credit_summary_item'):
            strings = tuple(tag.stripped_strings)
            if 'Creator:' in strings or 'Creators:' in strings:
                creators.extend(self._get_persons(tag))
        return tuple(creators)

    async def directors(self, id):
        soup = await self._get_soup(f'title/{id}')
        directors = []
        for tag in soup.find_all(class_='credit_summary_item'):
            strings = tuple(tag.stripped_strings)
            if 'Director:' in strings or 'Directors:' in strings:
                directors.extend(self._get_persons(tag))
        return tuple(directors)

    async def keywords(self, id):
        soup = await self._get_soup(f'title/{id}')
        subtext_tag = soup.find(class_='subtext')
        if subtext_tag:
            genre_links = subtext_tag.find_all('a', href=re.compile(r'/search/title\?genres='))
            return tuple(
                link.string.lower().strip()
                for link in genre_links
            )
        return ()

    async def poster_url(self, id):
        soup = await self._get_soup(f'title/{id}')
        poster_tag = soup.find(class_='poster')
        if poster_tag:
            link_tag = poster_tag.find('a')
            if link_tag and link_tag.get('href'):
                soup = await self._get_soup(link_tag['href'])
                img_tags = [
                    tag
                    for tag in soup.find_all('img', class_=re.compile(r'^MediaViewerImagestyles__PortraitImage'))
                    if 'peek' not in tag['class']
                ]
                if img_tags:
                    url = img_tags[0].get('src')
                    if url:
                        return url

    rating_min = 0.0
    rating_max = 10.0

    async def rating(self, id):
        soup = await self._get_soup(f'title/{id}')
        rating_tag = soup.find(itemprop='ratingValue')
        if rating_tag:
            try:
                return float(rating_tag.string)
            except (ValueError, TypeError):
                pass

    async def summary(self, id):
        soup = await self._get_soup(f'title/{id}')
        storyline_tag = soup.find(id='titleStoryLine')
        try:
            summary_tag = storyline_tag.div.p.span
        except AttributeError:
            pass
        else:
            if summary_tag:
                return ''.join(summary_tag.strings).strip()
        return ''

    async def title_english(self, id, allow_empty=True):
        akas = await self._get_akas(id)
        original_title = akas.get('(original title)', '')
        for key, english_title in akas.items():
            for regex in self._english_akas_keys:
                if regex.search(key):
                    # _log.debug('Interesting English title: %r -> %r', key, english_title)
                    if not allow_empty:
                        # _log.debug('Forcing first match: %r', english_title)
                        return english_title
                    if not self._titles_are_similar(english_title, original_title):
                        # _log.debug('English title: %r', english_title)
                        # _log.debug('Original title: %r', original_title)
                        return english_title
                    # else:
                    #     _log.debug('Similar to original title %r: %r', original_title, english_title)
        return ''

    async def title_original(self, id):
        akas = await self._get_akas(id)
        original_title = akas.get('(original title)', '')
        english_title = await self.title_english(id, allow_empty=False)
        if original_title:
            if not self._titles_are_similar(original_title, english_title):
                # _log.debug('Original title: %r', original_title)
                # _log.debug('English title: %r', english_title)
                return original_title
            # else:
            #     _log.debug('Similar to English title %r: %r', english_title, original_title)
        else:
            # Default to getting title from link to main page
            # _log.debug('No original title in AKAs found')
            soup = await self._get_soup(f'title/{id}/releaseinfo')
            title_tag = soup.find(class_='subpage_title_block__right-column')
            if title_tag:
                a_tag = title_tag.find('a', href=re.compile(r'/title/tt'))
                if a_tag:
                    return a_tag.string

        _log.debug('Defaulting to English title: %r', english_title)
        return english_title

    def _titles_are_similar(self, a, b):
        """Whether normalized `a` contains normalized `b` or vice versa"""
        an = self._normalize_title(a)
        bn = self._normalize_title(b)
        return an and bn and (an in bn or bn in an)

    _normalize_title_translation = str.maketrans('', '', string.punctuation)

    def _normalize_title(self, title):
        """Return casefolded `title` without punctuation and deduplicated whitespace"""
        return ' '.join(title.translate(self._normalize_title_translation).casefold().split())

    _ignored_akas_keys = (
        re.compile(r'\(TV title\)$'),
        re.compile(r'\(alternative spelling\)$'),
        re.compile(r'\(dubbed version\)$'),
        re.compile(r'\(literal title\)$'),
        re.compile(r'\(original script title\)$'),
        re.compile(r'\(short title\)$'),
        re.compile(r'\(video box title\)$'),
        re.compile(r'\(working title\)$'),
    )
    _english_akas_keys = (
        re.compile(r'^USA.*English'),
        re.compile(r'^USA$'),
        re.compile(r'^World-wide.*English'),
    )

    async def _get_akas(self, id):
        soup = await self._get_soup(f'title/{id}/releaseinfo')
        akas = {}

        def is_item_key_class(class_):
            # Class may also be named "aka-item__name-empty"
            return class_ and class_.startswith('aka-item__name')

        for item in soup.find_all('tr', class_='aka-item'):
            key_tag = item.find('td', class_=is_item_key_class)
            key = ''.join(key_tag.stripped_strings).strip()

            title_tag = item.find('td', class_='aka-item__title')
            title = ''.join(title_tag.stripped_strings).strip()

            if title:
                if not any(regex.search(key) for regex in self._ignored_akas_keys):
                    akas[key] = title
            #     else:
            #         _log.debug('Ignoring AKA: %r -> %r', key, title)
            # else:
            #     _log.debug('Ignoring empty title: %r -> %r', key, title)

        return akas

    async def type(self, id):
        soup = await self._get_soup(f'title/{id}')
        subtext_tag = soup.find(class_='subtext')
        if subtext_tag:
            # reversed() because interesting info is on the right side
            subtexts = ''.join(subtext_tag.stripped_strings).lower().split('|')
            genres = [genre.string.lower() for genre in subtext_tag.find_all(
                'a', href=re.compile(r'/search/title\?genres='))]
            if 'episode' in subtexts[-1]:
                return ReleaseType.episode
            elif 'tv series' in subtexts[-1]:
                return ReleaseType.season
            elif 'tv mini-series' in subtexts[-1]:
                return ReleaseType.season
            elif 'tv short' in subtexts[-1]:
                return ReleaseType.movie
            elif 'short' in genres:
                return ReleaseType.movie
            elif 'video' in subtexts[-1].split(' '):
                return ReleaseType.movie
            elif 'tv movie' in subtexts[-1]:
                return ReleaseType.movie
            elif re.search(r'^\d+ [a-z]+ \d{4}', subtexts[-1]):
                return ReleaseType.movie
        return ReleaseType.unknown

    async def url(self, id):
        return f'{self._url_base.rstrip("/")}/title/{id}'

    async def year(self, id):
        soup = await self._get_soup(f'title/{id}')

        # Movies
        year_tag = soup.find(id='titleYear')
        if year_tag:
            return ''.join(year_tag.stripped_strings).strip('()')

        # Series
        subtext_tag = soup.find(class_='subtext')
        if subtext_tag:
            # reversed() because year is on the very right
            for tag in reversed(subtext_tag.find_all()):
                text = ''.join(tag.stripped_strings)
                # Looking for one of these:
                #   - "TV Mini-Series (<YEAR>)"
                #   - "TV Series (<YEAR>–<year>)" (the "–" is an EN DASH / U+2013)
                #   - "<day> <month> <YEAR>"
                match = re.search(r'\b(\d{4})\b', text)
                if match:
                    return match.group(1)

        return ''


class _ImdbSearchResult(common.SearchResult):
    def __init__(self, *, soup, imdb_api):
        id = self._get_id(soup)
        return super().__init__(
            cast=self._get_cast(soup),
            countries=functools.partial(imdb_api.countries, id),
            director=self._get_director(soup),
            id=self._get_id(soup),
            keywords=self._get_keywords(soup),
            summary=self._get_summary(soup),
            title=self._get_title(soup),
            title_english=functools.partial(imdb_api.title_english, id),
            title_original=functools.partial(imdb_api.title_original, id),
            type=self._get_type(soup),
            url=self._get_url(soup),
            year=self._get_year(soup),
        )

    def _get_cast(self, soup):
        people = soup.find(string=re.compile(r'Stars?.*'))
        if people:
            names = tuple(name.string.strip() for name in people.parent.find_all('a'))
            return names[1:]
        else:
            return ()

    def _get_id(self, soup):
        href = soup.find('a').get('href')
        return re.sub(r'^.*/([t0-9]+)/.*$', r'\1', href)

    def _get_director(self, soup):
        people = soup.find(string=re.compile(r'Director?.*'))
        if people:
            director = people.parent.find('a')
            if director:
                return director.string.strip()
        else:
            return ''

    def _get_keywords(self, soup):
        try:
            keywords = soup.find(class_='genre').string.strip()
        except AttributeError:
            keywords = ''
        return tuple(kw.strip().casefold() for kw in keywords.split(','))

    def _get_summary(self, soup):
        tags = soup.find_all(class_='text-muted')
        summary = (tags[2].string or '').strip()

        # Look for "See full summary" link. Preceding text is summary.
        if not summary:
            summary_link = soup.find('a', text=re.compile(r'(?i:full\s+summary)'))
            if summary_link:
                summary_tag = summary_link.parent
                if summary_tag:
                    summary = ''.join(summary_tag.strings)
                    summary = re.sub(r'See full summary.*', '', summary).strip()
        return summary

    def _get_title(self, soup):
        return soup.find('a').string.strip()

    def _get_type(self, soup):
        if soup.find(string=re.compile(r'Directors?:')):
            return ReleaseType.movie
        else:
            return ReleaseType.series

    def _get_url(self, soup):
        id = self._get_id(soup)
        return f'{ImdbApi._url_base}/title/{id}'

    def _get_year(self, soup):
        try:
            year = soup.find(class_='lister-item-year').string or ''
        except AttributeError:
            return ''
        # Year may be preceded by roman number
        year = re.sub(r'\([IVXLCDM]+\)\s*', '', year)
        # Remove parentheses
        year = year.strip('()')
        # Possible formats:
        # - YYYY
        # - YYYY–YYYY  ("–" is NOT "U+002D / HYPHEN-MINUS" but "U+2013 / EN DASH")
        # - YYYY–
        try:
            return str(int(year[:4]))
        except (ValueError, TypeError):
            return ''
